/*
 * Scale Stack
 *
 * Copyright 2010 Eric Day
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * @file
 * @brief Echo Flood Datagram Definitions
 */

#include "config.h"

#include <cstring>

#include <scalestack/echo/flood/datagram.h>
#include <scalestack/echo/flood/flood.h>
#include <scalestack/kernel/core.h>
#include <scalestack/kernel/module.h>
#include <scalestack/kernel/option.h>

namespace scalestack
{
namespace echo
{
namespace flood
{

/*
 * Public methods.
 */

datagram::datagram(network::datagram_service& creator):
  network::datagram(creator),
  _iterations(_module.get_option("iterations").get_size_value()),
  _packet_size(_module.get_option("packet_size").get_size_value()),
  _read_offset(),
  _write_offset()
{
}

datagram::~datagram()
{
}

void datagram::ready(void)
{
  flush_send();
}

size_t datagram::receive(uint8_t* buffer, size_t size, struct sockaddr&, socklen_t)
{
  if (memcmp(random_buffer + _read_offset, buffer, size))
    _module.log_fatal(_("Received data did not match what was sent"));

  _read_offset += size;

  if (_read_offset >= random_buffer_size)
  {
    if (_iterations == 0 || --_iterations > 0)
    {
      _read_offset = 0;
      _write_offset = 0;

      flush_send();
    }
    else
      remove();
  }
  else
    flush_send();

  return size;
}

void datagram::flush_send(void)
{
  size_t packet_size;

  if (random_buffer_size - _write_offset < _packet_size)
    packet_size = random_buffer_size - _write_offset;
  else
    packet_size = _packet_size;

  _write_offset += send(random_buffer + _write_offset, packet_size);
}

void datagram::error(void)
{
  _module.log_error("Datagram error");
  _module.get_core().shutdown();
  remove();
}

} /* namespace flood */
} /* namespace echo */
} /* namespace scalestack */
