/*
 * Scale Stack
 *
 * Copyright 2010 Eric Day
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * @file
 * @brief Event Handler and Handler Provider Declarations
 */

#ifndef SCALESTACK_EVENT_HANDLER_H
#define SCALESTACK_EVENT_HANDLER_H

#include <ctime>

#include <scalestack/event/handler_service.h>
#include <scalestack/kernel/macros.h>

namespace scalestack
{
namespace event
{

class handler_provider;

/**
 * This is the event handler base class. Other classes that want to add and
 * handle events must derive from this.
 */
class SCALESTACK_API handler
{
public:

  handler(handler_service& handler_service);

  virtual ~handler();

  /**
   * Get the handler service this handler belongs to.
   */
  handler_service& get_handler_service(void);

  /**
   * This method is run after the event service has added it.
   */
  virtual void start(void);

  /**
   * Mark that this event handler should be removed.
   */
  void remove(void);

  /**
   * This method is run when the handler encounters an error. The default
   * behavior is to call remove().
   */
  virtual void error(void);

  /**
   * Set the timer.
   *
   * @param[in] seconds The number of seconds from now for the timer to
   *  expire. A value of 0 cancels the timer.
   */
  void set_timer(time_t seconds);

  /**
   * Run when the timer expires.
   */
  virtual void timer_expired(void);

  /**
   * Set the file descriptor to watch.
   *
   * @param[in] file_descriptor The file descriptor to use when watching
   *  for read and write events.
   */
  void set_file_descriptor(int file_descriptor);

  /**
   * Watch the file descriptor for reading.
   */
  void watch_read(void);

  /**
   * Run when the file descriptor is ready for reading.
   */
  virtual void read_ready(int file_descriptor);

  /**
   * Watch the file descriptor for writing.
   */
  void watch_write(void);

  /**
   * Run when the file descriptor is ready for writing.
   */
  virtual void write_ready(int file_descriptor);

  /**
   * Notify the event service to run this handler now.
   */
  void run_now(void);

  /**
   * Run when the event service receives a run_now() request.
   */
  virtual void run(void);

protected:

  handler_service& _handler_service;
  kernel::module& _module;

private:

  /**
   * Don't allow copying of objects.
   */
  SCALESTACK_LOCAL
  handler(const handler&);

  /**
   * Don't allow assignment of objects.
   */
  SCALESTACK_LOCAL
  handler& operator=(const handler&);

  handler_provider* _handler_provider;

  friend class handler_provider;
};

/**
 * This is the base class for classes that provide event handlers. Instances
 * of this class are bound to handler instances in order to provider the
 * implementation.
 */
class SCALESTACK_API handler_provider
{
public:

  handler_provider(void);

  virtual ~handler_provider();

  /**
   * Get the handler bound to this provider.
   */
  handler* get_handler(void) const;

  /**
   * Set the handler for this handler_provider. This is used while
   * initializing the event handler in the handler service class. This is
   * required in order to reuse both handler and handler_provider instances,
   * which will be implemented by unrelated classes.
   *
   * @param[in] handler Handler to bind to this provider.
   */
  void set_handler(handler* handler);

  /**
   * See handler::start().
   */
  void start(void);

  /**
   * See handler::remove().
   */
  virtual void remove(void) = 0;

  /**
   * See handler::error().
   */
  void error(void);

  /**
   * See handler::set_timer().
   */
  virtual void set_timer(time_t seconds) = 0;

  /**
   * See handler::timer_expired().
   */
  void timer_expired(void);

  /**
   * See handler::set_file_descriptor().
   */
  virtual void set_file_descriptor(int file_descriptor) = 0;

  /**
   * See handler::watch_read().
   */
  virtual void watch_read(void) = 0;

  /**
   * See handler::read_ready().
   */
  void read_ready(int file_descriptor);

  /**
   * See handler::watch_write().
   */
  virtual void watch_write(void) = 0;

  /**
   * See handler::write_ready().
   */
  void write_ready(int file_descriptor);

  /**
   * See handler::run_now().
   */
  virtual void run_now(void) = 0;

  /**
   * See handler::run().
   */
  void run(void);

private:

  /**
   * Don't allow copying of objects.
   */
  SCALESTACK_LOCAL
  handler_provider(const handler_provider&);

  /**
   * Don't allow assignment of objects.
   */
  SCALESTACK_LOCAL
  handler_provider& operator=(const handler_provider&);

  handler* _handler;
};

/*
 * Public handler methods.
 */

inline handler::handler(handler_service& handler_service):
  _handler_service(handler_service),
  _module(handler_service.get_module()),
  _handler_provider(NULL)
{
}

inline handler::~handler()
{
}

inline handler_service& handler::get_handler_service(void)
{
  return _handler_service;
}

inline void handler::start(void)
{
}

inline void handler::remove(void)
{
  _handler_provider->remove();
}

inline void handler::error(void)
{
  remove();
}

inline void handler::set_timer(time_t timer)
{
  _handler_provider->set_timer(timer);
}

inline void handler::timer_expired(void)
{
}

inline void handler::set_file_descriptor(int file_descriptor)
{
  _handler_provider->set_file_descriptor(file_descriptor);
}

inline void handler::watch_read(void)
{
  _handler_provider->watch_read();
}

inline void handler::read_ready(int)
{
}

inline void handler::watch_write(void)
{
  _handler_provider->watch_write();
}

inline void handler::write_ready(int)
{
}

inline void handler::run_now(void)
{
  _handler_provider->run_now();
}

inline void handler::run(void)
{
}

/*
 * Public handler_provider methods.
 */

inline handler_provider::handler_provider(void):
  _handler(NULL)
{
}

inline handler_provider::~handler_provider()
{
}

inline handler* handler_provider::get_handler(void) const
{
  return _handler;
}

inline void handler_provider::set_handler(handler* handler)
{
  _handler = handler;
  _handler->_handler_provider = this;
}

inline void handler_provider::start(void)
{
  _handler->start();
}

inline void handler_provider::error(void)
{
  _handler->error();
}

inline void handler_provider::timer_expired(void)
{
  _handler->timer_expired();
}

inline void handler_provider::read_ready(int file_descriptor)
{
  _handler->read_ready(file_descriptor);
}

inline void handler_provider::write_ready(int file_descriptor)
{
  _handler->write_ready(file_descriptor);
}

inline void handler_provider::run(void)
{
  _handler->run();
}

} /* namespace event */
} /* namespace scalestack */

#endif /* SCALESTACK_EVENT_HANDLER_H */
