/*
 * Scale Stack
 *
 * Copyright 2010 Eric Day
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * @file
 * @brief Event Handler Service and Handler Service Provider Declarations
 */

#ifndef SCALESTACK_EVENT_HANDLER_SERVICE_H
#define SCALESTACK_EVENT_HANDLER_SERVICE_H

#include <scalestack/event/service.h>
#include <scalestack/kernel/core.h>
#include <scalestack/kernel/macros.h>
#include <scalestack/kernel/module.h>
#include <scalestack/kernel/service.h>

namespace scalestack
{
namespace event
{

class handler;
class handler_service_provider;

/**
 * This is a base class that other classes using handler classes must
 * implement. This provides a way to add and remove event handlers from the
 * event service provider.
 */
class SCALESTACK_API handler_service: public kernel::service
{
public:

  /**
   * Constructor for event handler service.
   *
   * @param[in] module Module that the handler service is associated with.
   */
  handler_service(kernel::module& module);

  virtual ~handler_service();

  /**
   * Add an event handler to the handler service.
   *
   * @param[in] handler Event handler to add.
   */
  void add_handler(handler* handler);

  /**
   * Remove an event handler. This is called by the event service when the
   * handler is no longer in use.
   *
   * @param[in] handler Event handler to remove.
   */
  virtual void remove_handler(handler* handler) = 0;

  /**
   * Shutdown the handler service by removing all associated handlers. This
   * can block and should most likely only be used during shutdown in derived
   * class destructors.
   */
  void shutdown(void);

private:

  /**
   * Don't allow copying of objects.
   */
  SCALESTACK_LOCAL
  handler_service(const handler_service&);

  /**
   * Don't allow assignment of objects.
   */
  SCALESTACK_LOCAL
  handler_service& operator=(const handler_service&);

  handler_service_provider& _handler_service_provider;
};

/**
 * This is a base class that event service provider modules must
 * implement. This is used to provide functionality for handler services.
 */
class SCALESTACK_API handler_service_provider: public kernel::service
{
public:

  handler_service_provider(kernel::module& module,
                           handler_service& handler_service);

  virtual ~handler_service_provider();

  /**
   * See handler_service::add_handler().
   */
  virtual void add_handler(handler* handler) = 0;

  /**
   * See handler_service::remove_handler().
   */
  virtual void remove_handler(handler* handler);

  /**
   * See handler_service::add_handler().
   */
  virtual void shutdown(void) = 0;

private:

  /**
   * Don't allow copying of objects.
   */
  SCALESTACK_LOCAL
  handler_service_provider(const handler_service_provider&);

  /**
   * Don't allow assignment of objects.
   */
  SCALESTACK_LOCAL
  handler_service_provider& operator=(const handler_service_provider&);

  handler_service& _handler_service;
};

/*
 * Public handler_service methods.
 */

inline handler_service::handler_service(kernel::module& module):
  kernel::service(module),
  _handler_service_provider(static_cast<event::service&>(module.get_core().get_service("event"))._add_handler_service(*this))
{
}

inline handler_service::~handler_service()
{
  static_cast<event::service&>(_module.get_core().get_service("event"))._remove_handler_service(_handler_service_provider);
}

inline void handler_service::add_handler(handler* handler)
{
  _handler_service_provider.add_handler(handler);
}

inline void handler_service::shutdown(void)
{
  _handler_service_provider.shutdown();
}

/*
 * Public handler_service_provider methods.
 */

inline handler_service_provider::handler_service_provider(kernel::module& module, handler_service& handler_service):
  kernel::service(module),
  _handler_service(handler_service)
{
}

inline handler_service_provider::~handler_service_provider()
{
}

inline void handler_service_provider::remove_handler(handler* handler)
{
  _handler_service.remove_handler(handler);
}

} /* namespace event */
} /* namespace scalestack */

#endif /* SCALESTACK_EVENT_HANDLER_SERVICE_H */
