/*
 * Scale Stack
 *
 * Copyright 2010 Eric Day
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * @file
 * @brief libevent Event Handler Provider Definitions
 */

#include "config.h"

#include <cerrno>
#include <fcntl.h>
#include <unistd.h>

#include <scalestack/event/libevent/handler_provider.h>
#include <scalestack/event/libevent/thread.h>
#include <scalestack/kernel/module.h>

using namespace std;

namespace scalestack
{
namespace event
{
namespace libevent
{

/*
 * Public functions.
 */

extern "C" void handler_provider_callback(int file_descriptor,
                                          short events,
                                          void* context)
{
  handler_provider* handler_provider = static_cast<libevent::handler_provider*>(context);
  handler_provider->_callback(file_descriptor, events);
}

/*
 * Public methods.
 */

handler_provider::handler_provider(kernel::module& module, handler_service_provider& handler_service_provider):
  _in_check_queue(false),
  _need_start(false),
  _need_update(false),
  _need_run(false),
  _need_remove(false),
  _is_processing(false),
  _is_set(false),
  _is_added(false),
  _watch_read(false),
  _set_watch_read(false),
  _watch_write(false),
  _set_watch_write(false),
  _file_descriptor_updated(false),
  _file_descriptor(-1),
  _expire_time(0),
  _set_expire_time(0),
  _module(module),
  _handler_service_provider(handler_service_provider),
  _thread(NULL),
  _event(),
  _timer()
{
}

handler_service_provider& handler_provider::get_handler_service_provider(void)
{
  return _handler_service_provider;
}

handler_provider::~handler_provider()
{
  set_thread(NULL);
}

void handler_provider::set_thread(thread* thread)
{
  _thread = thread;

  if (_thread == NULL)
  {
    if (_is_added)
    {
      event_del(&_event);
      _is_added = false;
    }
  }
  else
  {
    _need_start = true;
    if (!_in_check_queue)
      _thread->check_handler_provider(this);
  }
}

bool handler_provider::add_to_check_queue(void)
{
  if (_in_check_queue)
    return false;

  _module.log_info("[%X] Adding to check queue", this);
  _in_check_queue = true;
  return true;
}

void handler_provider::check(void)
{
  _module.log_info("[%X] Checking", this);

  _in_check_queue = false;
  _is_processing = true;

  if (_need_start)
  {
    _need_start = false;
    start();
  }

  if (_need_run)
  {
    _need_run = false;
    run();
  }

  _is_processing = false;
  _need_update = false;
  _update_event();
}

bool handler_provider::need_remove(void)
{
  return _need_remove;
}

void handler_provider::remove(void)
{
  _module.log_info("[%X] Removing", this);
  _need_remove = true;
  if (!_in_check_queue)
    _thread->check_handler_provider(this);
}

void handler_provider::set_timer(time_t seconds)
{
  if (seconds == 0)
    _expire_time = 0;
  else
    _expire_time = _thread->get_now().tv_sec + seconds;

  _send_update_notification();
}

void handler_provider::set_file_descriptor(int file_descriptor)
{
  _file_descriptor = file_descriptor;
  _file_descriptor_updated = true;

  if (_file_descriptor != -1)
  {
    int return_code = fcntl(_file_descriptor, F_GETFL, 0);
    if (return_code == -1)
    {
      close(_file_descriptor);
      _module.log_fatal("Failed fcntl: %d", errno);
    }

    return_code = fcntl(_file_descriptor, F_SETFL, return_code | O_NONBLOCK);
    if (return_code == -1)
    {
      close(_file_descriptor);
      _module.log_fatal("Failed fcntl: %d", errno);
    }
  }

  _send_update_notification();
}

void handler_provider::watch_read(void)
{
  _watch_read = true;
  _send_update_notification();
}

void handler_provider::watch_write(void)
{
  _watch_write = true;
  _send_update_notification();
}

void handler_provider::run_now(void)
{
  _need_run = true;
  if (!_in_check_queue)
    _thread->check_handler_provider(this);
}

/*
 * Private methods.
 */

void handler_provider::_send_update_notification(void)
{
  /*
   * We call update_event() after processing callbacks, so don't bother doing
   * anything now if we're already in callbacks. This may be due to recursion
   * or coincidence with another thread.
   */
  if (_is_processing)
    return;

  /* Otherwise, notify the thread that this handler has changed. */
  if (_set_expire_time > _expire_time ||
      _file_descriptor_updated ||
      _set_watch_read != _watch_read ||
      _set_watch_write != _watch_write)
  {
    _need_update = true;
    if (!_in_check_queue)
      _thread->check_handler_provider(this);
  }
}

void handler_provider::_update_event(void)
{
  _need_update = false;

  if (_need_start)
    return;

  if (_set_expire_time <= _expire_time &&
      !_file_descriptor_updated &&
      _set_watch_read == _watch_read &&
      _set_watch_write == _watch_write)
  {
    return;
  }

  if (_is_added)
  {
    event_del(&_event);
    _is_added = false;
  }

  if (!_is_set ||
      _file_descriptor_updated ||
      _set_watch_read != _watch_read ||
      _set_watch_write != _watch_write)
  {
    short events = EV_PERSIST;

    if (_file_descriptor != -1)
    {
      if (_watch_read)
        events |= EV_READ;
      if (_watch_write)
        events |= EV_WRITE;
    }

    event_set(&_event, _file_descriptor, events, handler_provider_callback, this);

    if (event_base_set(_thread->get_event_base(), &_event) == -1)
    {
      _is_set = false;
      _module.log_error("Failed event_base_set: -1");
      remove();
      return;
    }

    _is_set = true;
    _file_descriptor_updated = false;
    _set_watch_read = _watch_read;
    _set_watch_write = _watch_write;
  }

  int return_code;
  if (_expire_time > 0)
  {
    _timer.tv_sec = _expire_time;
    return_code = event_add(&_event, &_timer);
  }
  else if (_file_descriptor != -1)
    return_code = event_add(&_event, NULL);
  else
    return_code = 0;

  if (return_code == -1)
  {
    _is_added = false;
    _module.log_error("Failed event_add: -1");
    remove();
    return;
  }

  _is_added = true;
}

void handler_provider::_callback(int file_descriptor, short events)
{
  _is_processing = true;

  if (events & EV_TIMEOUT && !_need_remove)
  {
    _set_expire_time = 0;

    if (_thread->get_now().tv_sec >= _expire_time)
    {
      _module.log_debug("[%X] Timer expired", this);
      _expire_time = 0;
      timer_expired();
    }
  }

  if (events & EV_READ && !_need_remove)
  {
    _module.log_debug("[%X] Read ready", this);
    _watch_read = false;
    read_ready(file_descriptor);
  }

  if (events & EV_WRITE && !_need_remove)
  {
    _module.log_debug("[%X] Write ready", this);
    _watch_write = false;
    write_ready(file_descriptor);
  }

  _is_processing = false;

  _update_event();
}

} /* namespace libevent */
} /* namespace event */
} /* namespace scalestack */
