/*
 * Scale Stack
 *
 * Copyright 2010 Eric Day
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * @file
 * @brief libevent Event Handler Provider Declarations
 */

#ifndef SCALESTACK_EVENT_LIBEVENT_HANDLER_PROVIDER_H
#define SCALESTACK_EVENT_LIBEVENT_HANDLER_PROVIDER_H

#include <event.h>

#include <scalestack/event/handler.h>
#include <scalestack/kernel/macros.h>

namespace scalestack
{
namespace kernel
{

class module;

} /* namespace kernel */

namespace event
{
namespace libevent
{

class thread;

/**
 * Function passed to libevent to invoke handler_provider::_callback().
 */
extern "C" void handler_provider_callback(int file_descriptor,
                                          short events,
                                          void* context);

/**
 * This is the event handler provider implementation for libevent.
 */
class SCALESTACK_API handler_provider: public event::handler_provider
{
public:

  /**
   * Constructor for handler provider.
   *
   * @param[in] module Module that the handler is associated with.
   * @param[in] handler_service_provider Owner of this handler provider.
   */
  handler_provider(kernel::module& module,
                   handler_service_provider& handler_service_provider);

  ~handler_provider();

  /**
   * Get the handler service provider this handler provider belongs to.
   */
  handler_service_provider& get_handler_service_provider(void);

  /**
   * Set the thread this handler provider belongs to.
   *
   * @param[in] thread Thread to set.
   */
  void set_thread(thread* thread);

  /**
   * Ask if the the handler provider should be added to the check queue. If
   * so, also mark that it is now in the queue.
   */
  bool add_to_check_queue(void);

  /**
   * Check to see if handler provider needs to be started, ran, or updated.
   */
  void check(void);

  /**
   * Check to see if this handler provider should be removed.
   */
  bool need_remove(void);

  /**
   * See event::handler::remove().
   */
  void remove(void);

  /**
   * See event::handler::set_timer().
   */
  void set_timer(time_t seconds);

  /**
   * See event::handler::set_file_descriptor().
   */
  void set_file_descriptor(int file_descriptor);

  /**
   * See event::handler::watch_read().
   */
  void watch_read(void);

  /**
   * See event::handler::watch_write().
   */
  void watch_write(void);

  /**
   * See event::handler::run_now().
   */
  void run_now(void);

private:

  /**
   * Don't allow copying of objects.
   */
  SCALESTACK_LOCAL
  handler_provider(const handler_provider&);

  /**
   * Update the event handler structure with any changes.
   */
  SCALESTACK_LOCAL
  void _update_event(void);

  /**
   * Don't allow assignment of objects.
   */
  SCALESTACK_LOCAL
  handler_provider& operator=(const handler_provider&);

  /**
   * Notify handler provider that events should be updated.
   */
  SCALESTACK_LOCAL
  void _send_update_notification(void);

  /**
   * This is called when a libevent event notification happens.
   *
   * @param[in] file_descriptor File descriptor for the event.
   * @param[in] events Type of events that are ready.
   */
  SCALESTACK_LOCAL
  void _callback(int file_descriptor, short events);

  bool _in_check_queue;
  bool _need_start;
  bool _need_update;
  bool _need_run;
  bool _need_remove;
  bool _is_processing;
  bool _is_set;
  bool _is_added;
  bool _watch_read;
  bool _set_watch_read;
  bool _watch_write;
  bool _set_watch_write;
  int _file_descriptor_updated;
  int _file_descriptor;
  time_t _expire_time;
  time_t _set_expire_time;
  kernel::module& _module;
  handler_service_provider& _handler_service_provider;
  thread* _thread;
  struct event _event;
  struct timeval _timer;

  friend void handler_provider_callback(int file_descriptor,
                                        short events,
                                        void* context);
};

} /* namespace libevent */
} /* namespace event */
} /* namespace scalestack */

#endif /* SCALESTACK_EVENT_LIBEVENT_HANDLER_PROVIDER_H */
