/*
 * Scale Stack
 *
 * Copyright 2010 Eric Day
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * @file
 * @brief libevent Event Handler Service Provider Definitions
 */

#include "config.h"

#include <algorithm>
#include <cerrno>
#include <cstring>
#include <exception>
#include <functional>
#include <unistd.h>

#include <scalestack/event/handler.h>
#include <scalestack/event/libevent/handler_service_provider.h>
#include <scalestack/event/libevent/service.h>
#include <scalestack/kernel/core.h>
#include <scalestack/kernel/module.h>

using namespace std;

namespace scalestack
{
namespace event
{
namespace libevent
{

/*
 * Public methods.
 */

handler_service_provider::handler_service_provider(kernel::module& module, event::handler_service& handler_service):
  event::handler_service_provider(module, handler_service),
  _shutdown(false),
  _handler_count(),
  _service(static_cast<libevent::service&>(module.get_core().get_service("event"))),
  _handler_mutex(),
  _notification_pipe(),
  _handlers()
{
  int return_code = pthread_mutex_init(&_handler_mutex, NULL);
  if (return_code != 0)
  {
    _module.log_fatal(_("Failed pthread_mutex_init: %s:%d"),
      strerror(return_code), return_code);
  }

  if (pipe(_notification_pipe) == -1)
  {
    _module.log_fatal(_("Could not create notification pipe: %s:%d"),
      strerror(errno), errno);
  }
}

handler_service_provider::~handler_service_provider()
{
  close(_notification_pipe[0]);
  close(_notification_pipe[1]);

  int return_code = pthread_mutex_destroy(&_handler_mutex);
  if (return_code != 0)
  {
    _module.log_error(_("Failed pthread_mutex_destroy: %s:%d"),
                      strerror(return_code), return_code);
  }
}

void handler_service_provider::add_handler(handler* handler)
{
  _service.add_handler(handler, *this);

  int return_code = pthread_mutex_lock(&_handler_mutex);
  if (return_code != 0)
  {
    _module.log_fatal(_("Failed pthread_mutex_lock: %s:%d"),
                      strerror(return_code), return_code);
  }

  try
  {
    _handlers.insert(handler);
    ++_handler_count;
  }
  catch (exception&)
  {
    return_code = pthread_mutex_unlock(&_handler_mutex);
    if (return_code != 0)
    {
      _module.log_error(_("Failed pthread_mutex_unlock: %s:%d"),
                        strerror(return_code), return_code);
    }

    throw;
  }

  return_code = pthread_mutex_unlock(&_handler_mutex);
  if (return_code != 0)
  {
    _module.log_fatal(_("Failed pthread_mutex_unlock: %s:%d"),
                      strerror(return_code), return_code);
  }

  if (_shutdown)
    handler->remove();
}

void handler_service_provider::remove_handler(handler* handler)
{
  int return_code = pthread_mutex_lock(&_handler_mutex);
  if (return_code != 0)
  {
    _module.log_fatal(_("Failed pthread_mutex_lock: %s:%d"),
                      strerror(return_code), return_code);
  }

  _handlers.erase(handler);

  return_code = pthread_mutex_unlock(&_handler_mutex);
  if (return_code != 0)
  {
    _module.log_fatal(_("Failed pthread_mutex_unlock: %s:%d"),
                      strerror(return_code), return_code);
  }

  event::handler_service_provider::remove_handler(handler);

  /* TODO Replace this lock/decrement with an atomic op at some point. */
  return_code = pthread_mutex_lock(&_handler_mutex);
  if (return_code != 0)
  {
    _module.log_fatal(_("Failed pthread_mutex_lock: %s:%d"),
                      strerror(return_code), return_code);
  }

  --_handler_count;

  if (_shutdown)
  {
    if (write(_notification_pipe[1], "\0", 1) == -1 && errno != EINTR)
    {
      _module.log_fatal(_("Could not write to notification pipe: %s:%d"),
        strerror(errno), errno);
    }
  }

  return_code = pthread_mutex_unlock(&_handler_mutex);
  if (return_code != 0)
  {
    _module.log_fatal(_("Failed pthread_mutex_unlock: %s:%d"),
                      strerror(return_code), return_code);
  }
}

void handler_service_provider::shutdown(void)
{
  _shutdown = true;

  int return_code = pthread_mutex_lock(&_handler_mutex);
  if (return_code != 0)
  {
    _module.log_fatal(_("Failed pthread_mutex_lock: %s:%d"),
                      strerror(return_code), return_code);
  }

  for_each(_handlers.begin(), _handlers.end(), mem_fun(&handler::remove));

  char buffer[1024];
  while (_handler_count > 0)
  {
    return_code = pthread_mutex_unlock(&_handler_mutex);
    if (return_code != 0)
    {
      _module.log_fatal(_("Failed pthread_mutex_unlock: %s:%d"),
                        strerror(return_code), return_code);
    }

    if (read(_notification_pipe[0], buffer, sizeof(buffer)) == -1 &&
      errno != EINTR)
    {
      _module.log_fatal(_("Could not read from notification pipe: %s:%d"),
        strerror(errno), errno);
    }

    return_code = pthread_mutex_lock(&_handler_mutex);
    if (return_code != 0)
    {
      _module.log_fatal(_("Failed pthread_mutex_lock: %s:%d"),
                        strerror(return_code), return_code);
    }
  }

  return_code = pthread_mutex_unlock(&_handler_mutex);
  if (return_code != 0)
  {
    _module.log_fatal(_("Failed pthread_mutex_unlock: %s:%d"),
                      strerror(return_code), return_code);
  }
}

} /* namespace libevent */
} /* namespace event */
} /* namespace scalestack */
