/*
 * Scale Stack
 *
 * Copyright 2010 Eric Day
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * @file
 * @brief libevent Event Service Provider Definitions
 */

#include "config.h"

#include <algorithm>
#include <exception>
#include <functional>
#include <memory>

#include <scalestack/event/libevent/handler_provider.h>
#include <scalestack/event/libevent/handler_service_provider.h>
#include <scalestack/event/libevent/service.h>
#include <scalestack/event/libevent/thread.h>
#include <scalestack/kernel/module.h>
#include <scalestack/kernel/option.h>

using namespace std;

namespace scalestack
{
namespace event
{
namespace libevent
{

/*
 * Public methods.
 */

service::service(kernel::module& module):
  event::service(module),
  _current_thread(),
  _threads()
{
  size_t thread_count = _module.get_option("threads").get_size_value();
  if (thread_count == 0)
    _module.log_fatal("The 'threads' option must be greater than 0");

  for (size_t thread_index = 0; thread_index < thread_count; thread_index++)
  {
    thread* thread = NULL;
    try
    {
      thread = new libevent::thread(*this, thread_index);
      _threads.push_back(thread);
    }
    catch (exception&)
    {
      if (thread != NULL)
      {
        thread->stop();
        delete thread;
      }

      _remove_threads();
      throw;
    }
  }
}

service::~service()
{
  _remove_threads();
}

void service::add_handler(event::handler* handler,
                          handler_service_provider& handler_service_provider)
{
  auto_ptr<handler_provider> handler_provider(new libevent::handler_provider(_module, handler_service_provider));

  handler_provider->set_handler(handler);

  /* Choose assigned thread by round-robin. */
  _current_thread++;
  handler_provider->set_thread(_threads[_current_thread % _threads.size()]);

  handler_provider.release();
}

void service::remove_handler(handler_provider* handler_provider)
{
  handler_provider->get_handler_service_provider().remove_handler(handler_provider->get_handler());
  delete handler_provider;
}

/*
 * Private methods.
 */

void service::_remove_threads(void)
{
  for_each(_threads.begin(), _threads.end(), mem_fun(&thread::stop));

  for (threads::iterator thread = _threads.begin();
       thread != _threads.end();
       ++thread)
  {
    delete *thread;
  }
}

event::handler_service_provider& service::_add_handler_service(event::handler_service& handler_service)
{
  return *(new handler_service_provider(_module, handler_service));
}

void service::_remove_handler_service(event::handler_service_provider& handler_service_provider)
{
  delete &handler_service_provider;
}

} /* namespace libevent */
} /* namespace event */
} /* namespace scalestack */
