/*
 * Scale Stack
 *
 * Copyright 2010 Eric Day
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * @file
 * @brief libevent Event Service Declarations
 */

#ifndef SCALESTACK_EVENT_LIBEVENT_SERVICE_H
#define SCALESTACK_EVENT_LIBEVENT_SERVICE_H

#include <vector>

#include <scalestack/event/service.h>
#include <scalestack/kernel/macros.h>

namespace scalestack
{
namespace event
{

class handler;

namespace libevent
{

class handler_provider;
class thread;

class SCALESTACK_API service: public event::service
{
public:

  /**
   * Add the number of threads specified in the "threads" option.
   *
   * @param[in] module Module that the service is associated with.
   */
  service(kernel::module& module);

  ~service();

  /**
   * Add a handler to one of the threads in the service pool.
   *
   * @param[in] handler Handler to add.
   * @param[in] handler_service_provider Handler service the handler
   *  belongs to.
   */
  void add_handler(event::handler* handler,
                   handler_service_provider& handler_service_provider);

  /**
   * Remove handler that was previously added in add_handler().
   *
   * @param[in] handler_provider Handler provider to remove.
   */
  void remove_handler(handler_provider* handler_provider);

private:

  /**
   * Don't allow copying of objects.
   */
  SCALESTACK_LOCAL
  service(const service&);

  /**
   * Don't allow assignment of objects.
   */
  SCALESTACK_LOCAL
  service& operator=(const service&);

  /**
   * Remove all threads that were created in the constructor.
   */
  void _remove_threads(void);

  /**
   * See event::service::_add_handler_service().
   */
  event::handler_service_provider& _add_handler_service(event::handler_service& handler_service);

  /**
   * See event::service::_remove_handler_service().
   */
  void _remove_handler_service(event::handler_service_provider& handler_service_provider);

  typedef std::vector<thread*> threads;

  size_t _current_thread;
  threads _threads;
};

} /* namespace libevent */
} /* namespace event */
} /* namespace scalestack */

#endif /* SCALESTACK_EVENT_LIBEVENT_SERVICE_H */
