/*
 * Scale Stack
 *
 * Copyright 2010 Eric Day
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * @file
 * @brief libevent Event Thread Definitions
 */

#include "config.h"

#include <algorithm>
#include <cerrno>
#include <cstring>
#include <fcntl.h>
#include <functional>
#include <unistd.h>

#include <scalestack/event/libevent/handler_provider.h>
#include <scalestack/event/libevent/service.h>
#include <scalestack/event/libevent/thread.h>
#include <scalestack/kernel/core.h>
#include <scalestack/kernel/module.h>
#include <scalestack/kernel/option.h>

using namespace std;

namespace scalestack
{
namespace event
{
namespace libevent
{

/*
 * Public functions.
 */

extern "C" void* thread_run(void* context)
{
  thread* thread = static_cast<libevent::thread*>(context);
  thread->_run();
  return NULL;
}

extern "C" void thread_handle_notification(int file_descriptor, short events,
                                           void* context)
{
  thread* thread = static_cast<libevent::thread*>(context);
  thread->_handle_notification(file_descriptor, events);
}

/*
 * Public methods.
 */

thread::thread(service& service, size_t thread_index):
  _shutdown(false),
  _handler_providers_index(),
  _thread_index(thread_index),
  _module(service.get_module()),
  _service(service),
  _base(),
  _notification_pipe(),
  _thread_id(),
  _notification_event(),
  _notification_timeout(),
  _now(),
  _handler_providers_mutex(),
  _handler_providers()
{
  if (gettimeofday(&_now, 0) == -1)
    _module.log_fatal("Failed gettimeofday: %s:%d", strerror(errno), errno);

  _base = event_base_new();
  if (_base == NULL)
    _module.log_fatal("Failed event_base_new: NULL");

  _module.log_debug("[Thread %u] Using libevent %s with %s", _thread_index,
                   event_get_version(), event_base_get_method(_base));

  try
  {
    _add_mutexes();

    try
    {
      _add_notification_pipe();

      try
      {
        _add_thread();
      }
      catch (exception&)
      {
        _remove_notification_pipe();
        throw;
      }
    }
    catch (exception&)
    {
      _remove_mutexes();
      throw;
    }
  }
  catch (exception&)
  {
    event_base_free(_base);
    throw;
  }

  _module.log_info("[Thread %u] Constructed", _thread_index);
}

thread::~thread()
{
  _remove_thread();
  _remove_notification_pipe();
  _remove_mutexes();
  event_base_free(_base);
  _module.log_info("[Thread %u] Destructed", _thread_index);
}

void thread::stop(void)
{
  _module.log_info("[Thread %u] Stopping", _thread_index);
  _shutdown = true;
  _send_notification();
}

void thread::check_handler_provider(handler_provider* handler_provider)
{
  int return_code = pthread_mutex_lock(&_handler_providers_mutex);
  if (return_code != 0)
  {
    _module.log_fatal("Failed pthread_mutex_lock: %s:%d",
                      strerror(return_code), return_code);
  }

  /* Double check queue flag inside lock to avoid race conditions. */
  if (handler_provider->add_to_check_queue())
    _handler_providers[_handler_providers_index].push_back(handler_provider);

  return_code = pthread_mutex_unlock(&_handler_providers_mutex);
  if (return_code != 0)
  {
    _module.log_fatal("Failed pthread_mutex_unlock: %s:%d",
                      strerror(return_code), return_code);
  }

  _send_notification();
}

struct event_base* thread::get_event_base(void)
{
  return _base;
}

const struct timeval& thread::get_now(void)
{
  return _now;
}

/*
 * Private methods.
 */

void thread::_run(void)
{
  _module.log_info("[Thread %u] Running", _thread_index);

  if (event_base_loop(_base, 0) == -1)
  {
    _module.log_error("[Thread %u] event_base_loop: -1", _thread_index);
    _module.get_core().shutdown();
  }

  _module.log_info("[Thread %u] Stopped", _thread_index);
}

void thread::_add_notification_pipe(void)
{
  int return_code = pipe(_notification_pipe);
  if (return_code == -1)
    _module.log_fatal("Failed pipe: %s:%d", strerror(errno), errno);

  for (int index = 0; index < 2; index++)
  {
    return_code = fcntl(_notification_pipe[index], F_GETFL, 0);
    if (return_code == -1)
    {
      _remove_notification_pipe();
      _module.log_fatal("Failed fcntl: %s:%d", strerror(errno), errno);
    }

    return_code = fcntl(_notification_pipe[index], F_SETFL, return_code | O_NONBLOCK);
    if (return_code == -1)
    {
      _remove_notification_pipe();
      _module.log_fatal("Failed fcntl: %s:%d", strerror(errno), errno);
    }
  }

  event_set(&_notification_event, _notification_pipe[0], EV_READ | EV_PERSIST,
            thread_handle_notification, this);

  if (event_base_set(_base, &_notification_event) == -1)
  {
    _remove_notification_pipe();
    _module.log_fatal("Failed event_base_set: -1");
  }

  _notification_timeout.tv_sec = 1;
  _notification_timeout.tv_usec = 0;
  if (event_add(&_notification_event, &_notification_timeout) == -1)
  {
    _remove_notification_pipe();
    _module.log_fatal("Failed event_add: -1");
  }
}

void thread::_remove_notification_pipe(void)
{
  close(_notification_pipe[0]);
  close(_notification_pipe[1]);
}

void thread::_send_notification(void)
{
  if (write(_notification_pipe[1], "\0", 1) == -1)
  {
    _module.log_error("Could not write to notification pipe: %s:%d",
                      strerror(errno), errno);
  }
}

void thread::_handle_notification(int file_descriptor, short events)
{
  _module.log_debug("[Thread %u] Notified", _thread_index);

  if (events & EV_READ)
  {
    char buffer[1024];
    if (read(file_descriptor, buffer, sizeof(buffer)) == -1)
    {
      _module.log_error("Could not read from notification pipe: %s:%d",
                        strerror(errno), errno);
    }
  }

  if (event_del(&_notification_event) == -1)
  {
    /* Not much we can really do here, but this should never happen. */
    _module.log_error("Failed event_del: -1");
    _module.get_core().shutdown();
  }
  else if (!_shutdown)
  {
    if (event_add(&_notification_event, &_notification_timeout) == -1)
    {
      /* Not much we can really do here, but this should never happen. */
      _module.log_error("Failed event_add: -1");
      _module.get_core().shutdown();
    }
  }

  if (gettimeofday(&_now, 0) == -1)
    _module.log_error("Failed gettimeofday: %s:%d", strerror(errno), errno);

  /* Just switch the index, keep lock time to a minimum. */
  uint8_t index = _handler_providers_index;

  int return_code = pthread_mutex_lock(&_handler_providers_mutex);
  if (return_code != 0)
  {
    _module.log_fatal("Failed pthread_mutex_lock: %s:%d",
                      strerror(return_code), return_code);
  }

  _handler_providers_index = static_cast<uint8_t>(1 - _handler_providers_index);

  return_code = pthread_mutex_unlock(&_handler_providers_mutex);
  if (return_code != 0)
  {
    _module.log_fatal("Failed pthread_mutex_unlock: %s:%d",
                      strerror(return_code), return_code);
  }

  handler_providers::iterator handler_provider;
  for (handler_provider = _handler_providers[index].begin();
       handler_provider != _handler_providers[index].end();
       ++handler_provider)
  {
    if ((*handler_provider)->need_remove())
      _service.remove_handler(*handler_provider);
    else
      (*handler_provider)->check();
  }

  _handler_providers[index].clear();
}

void thread::_add_mutexes(void)
{
  int return_code = pthread_mutex_init(&_handler_providers_mutex, NULL);
  if (return_code != 0)
  {
    _module.log_fatal("Failed pthread_mutex_init: %s:%d",
                      strerror(return_code), return_code);
  }
}

void thread::_remove_mutexes(void)
{
  int return_code = pthread_mutex_destroy(&_handler_providers_mutex);
  if (return_code != 0)
  {
    _module.log_error("Failed pthread_mutex_destroy: %s:%d",
                      strerror(return_code), return_code);
  }
}

void thread::_add_thread(void)
{
  pthread_attr_t thread_attribute;

  int return_code = pthread_attr_init(&thread_attribute);
  if (return_code != 0)
  {
    _module.log_fatal("Failed pthread_attr_init: %s:%d",
                      strerror(return_code), return_code);
  }

  size_t stack_size = _module.get_option("thread_stack_size").get_size_value();
  if (stack_size != 0)
  {
    return_code = pthread_attr_setstacksize(&thread_attribute, stack_size);
    if (return_code != 0)
    {
      (void) pthread_attr_destroy(&thread_attribute);
      _module.log_fatal("Failed pthread_attr_setstacksize: %s:%d",
                      strerror(return_code), return_code);
    }
  }

  return_code = pthread_attr_setscope(&thread_attribute, PTHREAD_SCOPE_SYSTEM);
  if (return_code != 0)
  {
    (void) pthread_attr_destroy(&thread_attribute);
    _module.log_fatal("Failed pthread_attr_setscope: %s:%d",
                      strerror(return_code), return_code);
  }

  return_code = pthread_create(&_thread_id, &thread_attribute, thread_run, this);
  if (return_code != 0)
  {
    (void) pthread_attr_destroy(&thread_attribute);
    _module.log_fatal("Failed pthread_create: %s:%d",
                      strerror(return_code), return_code);
  }

  return_code = pthread_attr_destroy(&thread_attribute);
  if (return_code != 0)
  {
    _module.log_error("Failed pthread_attr_destroy: %s:%d",
                      strerror(return_code), return_code);
  }
}

void thread::_remove_thread(void)
{
  int return_code = pthread_join(_thread_id, NULL);
  if (return_code != 0)
  {
    _module.log_error("Failed pthread_join: %s:%d",
                      strerror(return_code), return_code);
  }
}

} /* namespace libevent */
} /* namespace event */
} /* namespace scalestack */
