/*
 * Scale Stack
 *
 * Copyright 2010 Eric Day
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * @file
 * @brief libevent Event Thread Declarations
 */

#ifndef SCALESTACK_EVENT_LIBEVENT_THREAD_H
#define SCALESTACK_EVENT_LIBEVENT_THREAD_H

#include <event.h>
#include <pthread.h>
#include <stdint.h>
#include <vector>

#include <scalestack/kernel/macros.h>

namespace scalestack
{
namespace kernel
{

class module;

} /* namespace kernel */

namespace event
{
namespace libevent
{

class handler_provider;
class service;

/**
 * Function passed to pthread_create() to invoke thread::_run().
 */
extern "C" void* thread_run(void* context);

/**
 * Function used by notification event to invoke thread::_handle_notification().
 */
extern "C" void thread_handle_notification(int file_descriptor,
                                           short events,
                                           void* context);

class thread
{
public:

  /**
   * Constructor for event threads. This also starts a new pthread. Once
   * starting this call, the _run() method can be called at any time, even
   * before the constructor completes.
   *
   * @param[in] service Service that is creating this thread.
   * @param[in] thread_index The 0-based index for this thread.
   */
  thread(service& service, size_t thread_index);

  ~thread();

  /**
   * Notify a thread to stop running. This is called by the service provider
   * class for each thread before deleting the threads. This allows all
   * threads to be notified to stop before any joins are attempted.
   */
  void stop(void);

  /**
   * Queue a handler provider to be checked on this thread.
   *
   * @param[in] handler_provider Event handler provider to check.
   */
  void check_handler_provider(handler_provider* handler_provider);

  /**
   * Get the event base struct. This is used by handler providers once they
   * are added to this thread.
   */
  struct event_base* get_event_base(void);

  /**
   * Get a cached time structure.
   */
  const struct timeval& get_now(void);

private:

  /**
   * Don't allow copying of objects.
   */
  SCALESTACK_LOCAL
  thread(const thread&);

  /**
   * Don't allow assignment of objects.
   */
  SCALESTACK_LOCAL
  thread& operator=(const thread&);

  /**
   * Method that implements the body of the running thread.
   */
  SCALESTACK_LOCAL
  void _run(void);

  /**
   * Start the notification pipe.
   */
  SCALESTACK_LOCAL
  void _add_notification_pipe(void);

  /**
   * Stop the notification pipe.
   */
  SCALESTACK_LOCAL
  void _remove_notification_pipe(void);

  /**
   * Send a notification to run and process events.
   */
  SCALESTACK_LOCAL
  void _send_notification(void);

  /**
   * Handle a notification event.
   */
  SCALESTACK_LOCAL
  void _handle_notification(int file_descriptor, short events);

  /**
   * Add mutexes needed for thread.
   */
  SCALESTACK_LOCAL
  void _add_mutexes(void);

  /**
   * Remove mutexes added in _add_mutexes().
   */
  SCALESTACK_LOCAL
  void _remove_mutexes(void);

  /**
   * Start the pthread.
   */
  SCALESTACK_LOCAL
  void _add_thread(void);

  /**
   * Stop the pthread by joining it.
   */
  SCALESTACK_LOCAL
  void _remove_thread(void);

  typedef std::vector<handler_provider*> handler_providers;

  bool _shutdown;
  uint8_t _handler_providers_index;
  size_t _thread_index;
  kernel::module& _module;
  service& _service;
  struct event_base* _base;
  int _notification_pipe[2];
  pthread_t _thread_id;
  struct event _notification_event;
  struct timeval _notification_timeout;
  struct timeval _now;
  pthread_mutex_t _handler_providers_mutex;
  handler_providers _handler_providers[2];

  friend void* thread_run(void*);
  friend void thread_handle_notification(int file_descriptor,
                                         short events,
                                         void* context);
};

} /* namespace libevent */
} /* namespace event */
} /* namespace scalestack */

#endif /* SCALESTACK_EVENT_LIBEVENT_THREAD_H */
