/*
 * Scale Stack
 *
 * Copyright 2010 Eric Day
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * @file
 * @brief Event Service Declarations
 */

#ifndef SCALESTACK_EVENT_SERVICE_H
#define SCALESTACK_EVENT_SERVICE_H

#include <scalestack/kernel/macros.h>
#include <scalestack/kernel/service.h>

namespace scalestack
{
namespace event
{

class handler_service;
class handler_service_provider;

/**
 * This is a base class that event providers must implement in order to
 * provide an event service.
 */
class SCALESTACK_API service: public kernel::service
{
public:

  /**
   * Constructor for event service.
   *
   * @param[in] module Module that the service is associated with.
   */
  service(kernel::module& module);

  virtual ~service();

private:

  /**
   * Don't allow copying of objects.
   */
  SCALESTACK_LOCAL
  service(const service&);

  /**
   * Don't allow assignment of objects.
   */
  SCALESTACK_LOCAL
  service& operator=(const service&);

  /**
   * Add a handler service to the event service.
   *
   * @param[in] handler_service Handler service to add.
   * @return Handler service provider to use with the handler service.
   */
  virtual handler_service_provider& _add_handler_service(handler_service& handler_service) = 0;

  /**
   * Remove a handler service provider from an event service.
   *
   * @param[in] handler_service_provider Handler service provider to remove.
   */
  virtual void _remove_handler_service(handler_service_provider& handler_service_provider) = 0;

  friend class handler_service;
};

/*
 * Public handler_service methods.
 */

inline service::service(kernel::module& module):
  kernel::service(module)
{
}

inline service::~service()
{
}

} /* namespace event */
} /* namespace scalestack */

#endif /* SCALESTACK_EVENT_SERVICE_H */
