/*
 * Scale Stack
 *
 * Copyright 2010 Eric Day
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * @file
 * @brief Kernel Core Declarations
 */

#ifndef SCALESTACK_KERNEL_CORE_H
#define SCALESTACK_KERNEL_CORE_H

#include <map>
#include <string>
#include <vector>

#include <scalestack/kernel/logger.h>
#include <scalestack/kernel/macros.h>

namespace scalestack
{
namespace kernel
{

class module;
class service;
class log;

/**
 * This class is the central part of the kernel and is responsible for
 * managing modules, services, and kernel logs.
 */
class SCALESTACK_API core: private logger
{
  friend class module;

public:

  /**
   * Initialize all data members and add a log_console object to the list
   * of logs.
   */
  core(void);

  /**
   * Ensure all modules are properly cleaned up and generate error messages
   * if there are leftover services or logs.
   */
  ~core();

  /* This is here to make the base method public. */
  using logger::set_threshold;

  /**
   * Parse arguments, most likely from the main() function arguments.
   *
   * @param[in] argument_count Argument count.
   * @param[in] arguments Array of argument values.
   */
  void parse_arguments(int argument_count, const char** arguments);

  /**
   * Parse arguments from a file. Comments start with a '#', and all empty
   * lines are ignored.
   *
   * @param[in] filename Name of the file to parse.
   */
  void parse_file(const std::string& filename);

  /**
   * Add an option value for a module. If a value already exists, the old
   * value is overwritten. Options can be of the form:
   *
   * module_name - Add the module.
   * module_name=value - Add a module alias from module_name to value.
   * module_name.option_name=value - Set a module option.
   *
   * @param[in] module_name Name of the module to add the option for.
   * @param[in] option_name Name of the option.
   * @param[in] value Value of the option.
   */
  void add_module_option_value(const std::string& module_name,
                               const std::string& option_name,
                               const std::string& value);

  /**
   * Set common signal handlers. Calling applications can either use this for
   * convenience or handle their own signals. SIGPIPE is set to be ignored,
   * and the following signals are set to call handle_signal(): SIGHUP,
   * SIGINT, SIGQUIT, SIGTERM, SIGUSR1, and SIGUSR2.
   */
  void set_signal_handlers(void);

  /**
   * Handle signals received by application signal handlers. This currently
   * process SIGINT, SIGQUIT, and SIGTERM as shutdown requests. This can be
   * used by applications that wish to handle their own signals.
   *
   * @param[in] signal_number Signal that was caught and needs to be handled.
   */
  void handle_signal(int signal_number);

  /**
   * Get a module that has been added to the core, or add one if a module
   * by that name doesn't exist.
   *
   * @param[in] name Name of the module to add. This is used in constructing
   *  the path of the module to load. Name comparisons are case insensitive.
   * @return The module that was added.
   */
  module& get_or_add_module(const std::string& name);

  /**
   * Get a module that has been added to the core.
   *
   * @param[in] name Name of the module to look for. This is used to check
   *  both the name and alias list. Name comparisons are case insensitive.
   * @return The module that was found.
   */
  module& get_module(const std::string& name) const;

  /**
   * Get a count of the number of modules in the core.
   */
  size_t get_module_count(void) const;

  /**
   * Reset the module iterator to the beginning and return first module. This
   * should be called first before iterating through all modules. The most
   * common way to use this interface is:
   *
   * @code
   *   while (module* current_module = core.begin_module();
   *          current_module != NULL;
   *          current_module = core.next_module())
   *   {
   *     // Do something with current_module here.
   *   }
   * @endcode
   *
   * @return First module or NULL if there are no modules.
   */
  module* begin_module(void);

  /**
   * Get the next module while iterating, see begin_module() for usage
   * details.
   *
   * @return Next module or NULL if at the end of the list.
   */
  module* next_module(void);

  /**
   * Add a service object for the given name. If a service by that name
   * already exists, this will throw an exception.
   *
   * @param[in] name Name of the service to add.
   * @param[in] service The service object to add.
   */
  void add_service(const std::string& name, service* service);

  /**
   * Remove a service object for the given name.
   *
   * @param[in] name Name of the service to remove.
   * @return The service object that was removed.
   */
  service* remove_service(const std::string& name);

  /**
   * Get a service object for the given name.
   *
   * @param[in] name Name of the service to get.
   * @return The service object that was requested.
   */
  service& get_service(const std::string& name) const;

  /**
   * Get a count of the number of services that have been added.
   */
  size_t get_service_count(void) const;

  /**
   * Add a log object for the given name. If a log by that name already
   * exists, this will throw an exception.
   *
   * @param[in] name Name of the log to add.
   * @param[in] log The log object to add.
   */
  void add_log(const std::string& name, log* log);

  /**
   * Remove a log object for the given name.
   *
   * @param[in] name Name of the log to remove.
   * @return The log object that was removed.
   */
  log* remove_log(const std::string& name);

  /**
   * Get a log object for the given name.
   *
   * @param[in] name Name of the log to get.
   * @return The log object that was requested.
   */
  log& get_log(const std::string& name) const;

  /**
   * Get a count of the number of logs that have been added.
   */
  size_t get_log_count(void) const;

  /**
   * This is the main loop of the core and will continue running until all
   * modules have been removed or a shutdown request is received.
   */
  void run(void);

  /**
   * Notify the main loop that it should wake up and process pending events.
   */
  void notify(void);

  /**
   * Notify the kernel core that it should start the shutdown process. This is
   * usually called from signal handlers or modules providing administrative
   * functions.
   */
  void shutdown(void);

private:

  /**
   * Don't allow copying of objects.
   */
  SCALESTACK_LOCAL
  core(const core&);

  /**
   * Don't allow assignment of objects.
   */
  SCALESTACK_LOCAL
  core& operator=(const core&);

  /**
   * Parse a single argument from any source. The format is
   * module[.option][=value], and the parsed value is passed into
   * add_module_option_value().
   *
   * @param[in] argument Argment to parse.
   */
  void _parse_argument(std::string argument);

  /**
   * Resolve module names using the alias map.
   *
   * @param[in] name Name to look up in alias map.
   * @return Alias if found, or original name that was passed in.
   */
  SCALESTACK_LOCAL
  const std::string& _get_alias(const std::string& name) const;

  /**
   * Get a module that has been added to the core, this is only used
   * internally. This behaves just like get_module() (and is used by that
   * method) but does not throw exceptions when the module is not found.
   */
  SCALESTACK_LOCAL
  module* _get_module(const std::string& name) const;

  /**
   * Set a fatal message as the error string. This is declared const so
   * other const methods can call it.
   *
   * @param[in] format Message format string.
   * @param[in] arguments Argument list for formatted message.
   */
  SCALESTACK_LOCAL
  void _set_fatal_message(const char* format, std::va_list arguments) const;

  /**
   * Log a core message. This is declared const so other const methods can
   * call it.
   *
   * @param[in] log_level Log level for this message.
   * @param[in] format Message format string.
   * @param[in] arguments Argument list for formatted message.
   */
  SCALESTACK_LOCAL
  void _log_message(log::level log_level,
                    const char* format,
                    std::va_list arguments) const;

  /**
   * Send a log message to all configured logs. This is declared const so
   * other const methods can call it.
   *
   * @param[in] log_level Log level for this message.
   * @param[in] context Context string for this message.
   * @param[in] format Message format string.
   * @param[in] arguments Argument list for formatted message.
   */
  SCALESTACK_LOCAL
  void _log(log::level log_level,
            const char* context,
            const char* format,
            std::va_list arguments) const;

  /**
   * Print a help message for the kernel and all loaded modules with options.
   */
  SCALESTACK_LOCAL
  void _print_help(void);

  /**
   * Print the parsed configuration options for the kernel and all modules.
   */
  SCALESTACK_LOCAL
  void _print_config(void);

  typedef std::vector<std::string> files;
  typedef std::vector<std::string> paths;
  typedef std::vector<module*> modules;
  typedef std::map<std::string, std::string> aliases;
  typedef std::map<std::string, service*> services;
  typedef std::map<std::string, log*> logs;

  bool _need_help;
  bool _need_config;
  bool _shutdown_request;
  bool _iterator_on_modules;
  int _notification_pipe[2];
  files _open_files;
  paths _paths;
  modules _modules;
  modules _new_modules;
  modules::iterator _module_iterator;
  aliases _aliases;
  services _services;
  logs _logs;
};

} /* namespace kernel */
} /* namespace scalestack */

#endif /* SCALESTACK_KERNEL_CORE_H */
