/*
 * Scale Stack
 *
 * Copyright 2010 Eric Day
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * @file
 * @brief Log Declarations
 */

#ifndef SCALESTACK_KERNEL_LOG_H
#define SCALESTACK_KERNEL_LOG_H

#include <scalestack/kernel/macros.h>

namespace scalestack
{
namespace kernel
{

/**
 * This is the base class for all kernel log implementations.
 */
class SCALESTACK_API log
{
public:

  /**
   * Available log levels.
   */
  enum level
  {
    LEVEL_FATAL,
    LEVEL_ERROR,
    LEVEL_NOTICE,
    LEVEL_INFO,
    LEVEL_DEBUG
  };

  log(void);

  virtual ~log();

  /**
   * Write a log entry, this must be supplied by derived classes.
   *
   * @param[in] log_level Log level for this message.
   * @param[in] context Context string for this message.
   * @param[in] message Message to log.
   */
  virtual void write_log(log::level log_level,
                         const char* context,
                         const char* message) = 0;

  /**
   * Get the string representation for a givel log level.
   */
  const char* get_level_string(log::level log_level);

private:

  /**
   * Don't allow copying of objects.
   */
  log(const log&);

  /**
   * Don't allow assignment of objects.
   */
  log& operator=(const log&);
};

/*
 * Public methods.
 */

inline log::log(void)
{
}

inline log::~log()
{
}

inline const char* log::get_level_string(log::level log_level)
{
  switch (log_level)
  {
  case LEVEL_FATAL:
    return "FATAL";
  case LEVEL_ERROR:
    return "ERROR";
  case LEVEL_NOTICE:
    return "NOTICE";
  case LEVEL_INFO:
    return "INFO";
  case LEVEL_DEBUG:
    return "DEBUG";
  default:
    return "(BAD)";
  };
}

} /* namespace kernel */
} /* namespace scalestack */

#endif /* SCALESTACK_KERNEL_LOG_H */
