/*
 * Scale Stack
 *
 * Copyright 2010 Eric Day
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * @file
 * @brief Kernel Logger Base Class
 */

#ifndef SCALESTACK_KERNEL_LOGGER_H
#define SCALESTACK_KERNEL_LOGGER_H

#include <cstdarg>

#include <scalestack/kernel/exception.h>
#include <scalestack/kernel/log.h>

namespace scalestack
{
namespace kernel
{

/**
 * Base class available for kernel classes that produce log messages.
 */
class SCALESTACK_API logger
{
public:

  logger(size_t threshold);

  virtual ~logger();

  /**
   * Set the threshold for logging messages.
   *
   * @param[in] threshold The new log threshold to set.
   */
  void set_threshold(size_t threshold);

  /**
   * Log a fatal message with the formatted string if log threshold is met.
   *
   * @param[in] format Message format string.
   */
  void log_fatal(const char* format, ...) const;

  /**
   * Log an error message with the formatted string if log threshold is met.
   *
   * @param[in] format Message format string.
   */
  void log_error(const char* format, ...) const;

  /**
   * Log a notice message with the formatted string if log threshold is met.
   *
   * @param[in] format Message format string.
   */
  void log_notice(const char* format, ...) const;

  /**
   * Log an info message with the formatted string if log threshold is met.
   *
   * @param[in] format Message format string.
   */
  void log_info(const char* format, ...) const;

  /**
   * Log a debug message with the formatted string if log threshold is met.
   *
   * @param[in] format Message format string.
   */
  void log_debug(const char* format, ...) const;

protected:

  size_t _threshold;

private:

  /**
   * Don't allow copying of objects.
   */
  logger(const logger&);

  /**
   * Don't allow assignment of objects.
   */
  logger& operator=(const logger&);

  /**
   * Set a fatal error message. This is used so fatal messages can be caught
   * for error strings even if logging threshold is not met.
   *
   * @param[in] format Message format string.
   * @param[in] arguments Argument list for formatted message.
   */
  virtual void _set_fatal_message(const char* format,
                                  std::va_list arguments) const = 0;

  /**
   * Log the message.
   *
   * @param[in] log_level Log level for this message.
   * @param[in] format Message format string.
   * @param[in] arguments Argument list for formatted message.
   */
  virtual void _log_message(log::level log_level,
                            const char* format,
                            std::va_list arguments) const = 0;
};

/*
 * Public methods.
 */

inline logger::logger(size_t threshold):
  _threshold(threshold)
{
}

inline logger::~logger()
{
}

inline void logger::set_threshold(size_t threshold)
{
  _threshold = threshold;
}

inline void logger::log_fatal(const char* format, ...) const
{
  std::va_list arguments;
  va_start(arguments, format);

  /*
   * We know this is going to trigger a catch because fatal messages throw
   * exceptions, so this is done so we can call va_end before passing on
   * the exception.
   */
  try
  {
    if (_threshold >= 1)
      _log_message(log::LEVEL_FATAL, format, arguments);
    else
      _set_fatal_message(format, arguments);
  }
  catch (kernel::exception&)
  {
    va_end(arguments);
    throw;
  }

  va_end(arguments);
}

inline void logger::log_error(const char* format, ...) const
{
  if (_threshold >= 1)
  {
    std::va_list arguments;
    va_start(arguments, format);
    _log_message(log::LEVEL_ERROR, format, arguments);
    va_end(arguments);
  }
}

inline void logger::log_notice(const char* format, ...) const
{
  if (_threshold >= 1)
  {
    std::va_list arguments;
    va_start(arguments, format);
    _log_message(log::LEVEL_NOTICE, format, arguments);
    va_end(arguments);
  }
}

inline void logger::log_info(const char* format, ...) const
{
  if (_threshold >= 2)
  {
    std::va_list arguments;
    va_start(arguments, format);
    _log_message(log::LEVEL_INFO, format, arguments);
    va_end(arguments);
  }
}

inline void logger::log_debug(const char* format, ...) const
{
  if (_threshold >= 3)
  {
    std::va_list arguments;
    va_start(arguments, format);
    _log_message(log::LEVEL_DEBUG, format, arguments);
    va_end(arguments);
  }
}

} /* namespace kernel */
} /* namespace scalestack */

#endif /* SCALESTACK_KERNEL_LOGGER_H */
