/*
 * Scale Stack
 *
 * Copyright 2010 Eric Day
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * @file
 * @brief Kernel Module Declarations
 */

#ifndef SCALESTACK_KERNEL_MODULE_H
#define SCALESTACK_KERNEL_MODULE_H

#include <stdint.h>
#include <string>
#include <vector>

#include <scalestack/kernel/logger.h>
#include <scalestack/kernel/macros.h>

namespace scalestack
{
namespace kernel
{

class core;
class option;

/**
 * This class represents a module (shared object) in the system. It is
 * responsible for loading, running, unloading, and managing options.
 */
class SCALESTACK_API module: public logger
{
public:

  /**
   * Struct that every module needs to expose globally so that the module
   * loader can find and load it. Modules should use the macros at the end
   * of this header file to define this struct.
   */
  typedef struct
  {
    uint32_t api_version;
    const char* name;
    const char* version;
    const char* author;
    const char* title;
    const char* license;
    const char* dependencies;
    void (*options)(module&);
    void (*start)(module&);
    void (*stop)(module&);
    void (*run)(module&);
    void (*reconfigure)(module&);
  } definition;

  ~module();

  /* This is here to make the base method public. */
  using logger::set_threshold;

  /**
   * Get the Core object that this module belongs to.
   */
  core& get_core(void) const;

  /**
   * Get the name of the module used for loading.
   */
  const std::string& get_name(void) const;

  /**
   * Get the name of the module after it has been normalized for comparison.
   */
  std::string get_normalized_name(void) const;

  /**
   * Get the full pathname of the module used for loading.
   */
  const std::string& get_pathname(void) const;

  /**
   * Get the version of the module.
   */
  const std::string& get_version(void) const;

  /**
   * Get the author of the module.
   */
  const std::string& get_author(void) const;

  /**
   * Get the title of the module.
   */
  const std::string& get_title(void) const;

  /**
   * Get the license of the module.
   */
  const std::string& get_license(void) const;

  /**
   * Get the list of module dependencies that must be loaded before this
   * module.
   */
  const std::string& get_dependencies(void) const;

  /**
   * Get a count of the modules in parent dependency list.
   */
  size_t get_parent_count(void) const;

  /**
   * This behaves just like core::begin_module(), but for parent dependencies
   * of this module.
   */
  module* begin_parent(void);

  /**
   * This behaves just like core::next_module(), but for parent dependencies
   * of this module.
   */
  module* next_parent(void);

  /**
   * Get a count of the modules in child dependency list.
   */
  size_t get_child_count(void) const;

  /**
   * This behaves just like core::begin_module(), but for child dependencies
   * of this module.
   */
  module* begin_child(void);

  /**
   * This behaves just like core::next_module(), but for child dependencies
   * of this module.
   */
  module* next_child(void);

  /**
   * Check if the module is currently loaded.
   */
  bool is_loaded(void) const;

  /**
   * Add an option to a module. This is usually only called in the "options"
   * callback function specified in the module definition. If an option
   * already exists, NULL is returned and an error is set in the module
   * to notify that a duplicate error message should be returned once the
   * "options" callback has completed. This prevents module writers from
   * having to check the return code of every add_option() call.
   *
   * Options can be added after the "options" callback, but they will not
   * show up in the help menu, will not be allowed from the command line,
   * and error checking must be done in the module.
   *
   * When displaying help on the command line, the format will be:
   *
   * --module_name.name=value_name - help (Default: default_value)
   *
   * @param[in] name Name of the option.
   * @param[in] help Help string for of the option.
   * @param[in] value_name The name of the value to use in the help.
   * @param[in] default_value The default value to use for the option.
   * @return Option that was added.
   */
  option& add_option(const std::string& name,
                     const std::string& help,
                     const std::string& value_name,
                     const std::string& default_value);

  /**
   * Get an option that has been added to the module.
   *
   * @param[in] name Name of the option to look for.
   * @return The option that was requested.
   */
  option& get_option(const std::string& name) const;

  /**
   * Get a count of the number of options in the module.
   */
  size_t get_option_count(void) const;

  /**
   * This behaves just like core::begin_module(), but for options of this
   * module.
   */
  option* begin_option(void);

  /**
   * This behaves just like core::next_module(), but for options of this
   * module.
   */
  option* next_option(void);

  /**
   * Mark for the module _run function to be called from the core loop.
   */
  void run(void);

  /**
   * Normalize a module name by replacing '::' separators with '_' and make
   * it all lowercase so it is suitable for comparison.
   *
   * @param[in] name Module name to normalize.
   * @return A string with the normalized module name.
   */
  static std::string normalize_name(const std::string& name);

private:

  /**
   * Construct a module instance. This is private because only Core, which
   * is listed as a friend, can construct modules.
   *
   * @param[in] core Core that created this module.
   * @param[in] name Name of module to create.
   * @param[in] threshold Logging threshold.
   */
  SCALESTACK_LOCAL
  module(core& core, const std::string& name, size_t threshold);

  /**
   * Don't allow copying of objects.
   */
  SCALESTACK_LOCAL
  module(const module&);

  /**
   * Don't allow assignment of objects.
   */
  SCALESTACK_LOCAL
  module& operator=(const module&);

  /**
   * Load a module by first detecting it's dependencies, loading those
   * modules, and then loading the requested module. This also includes
   * calling the option callback in the module once loaded to get option
   * information.
   */
  SCALESTACK_LOCAL
  void _load(void);

  /**
   * Try loading the dependency list from the module based on module path
   * list and module name.
   */
  SCALESTACK_LOCAL
  void _load_dependencies(void);

  /**
   * Try loading the module definition structure using the pathname generated
   * in _load_dependencies.
   */
  SCALESTACK_LOCAL
  void _load_definition(void);

  /**
   * Unload the module, unloading child dependencies if needed in the process.
   */
  SCALESTACK_LOCAL
  void _unload(void);

  /**
   * Call the module start function, starting all parent dependencies
   * if needed.
   */
  SCALESTACK_LOCAL
  void _start(void);

  /**
   * Call the module run function if needed.
   */
  SCALESTACK_LOCAL
  void _run(void);

  /**
   * Call the module reconfigure function if needed.
   */
  SCALESTACK_LOCAL
  void _reconfigure(void);

  /**
   * Open the shared object in pathname and map module definition symbol.
   *
   * @param[in] flags Flags to pass to dlopen().
   */
  SCALESTACK_LOCAL
  void _open_object(int flags);

  /**
   * Close the shared object associated with module.
   */
  SCALESTACK_LOCAL
  void _close_object(void);

  /**
   * Get an option that has been added to the core, this is only used
   * internally. This behaves just like get_option() (and is used by that
   * method) but does not throw exceptions when the option is not found.
   */
  SCALESTACK_LOCAL
  option* _get_option(const std::string& name) const;

  /**
   * Get an option that has been added to the module, or add one if an option
   * by that name doesn't exist. Used internally by other methods.
   *
   * @param name Name of the option to add.
   * @return The option that was requested.
   */
  SCALESTACK_LOCAL
  option& _get_or_add_option(const std::string& name);

  /**
   * Set an option value. If a value already exists, the old value is
   * overwritten.
   *
   * @param[in] name Name of the option.
   * @param[in] value Value of the option.
   */
  SCALESTACK_LOCAL
  void _set_option_value(const std::string& name, const std::string& value);

  /**
   * Check to see if all options set for the module are valid.
   */
  SCALESTACK_LOCAL
  void _check_options(void) const;

  /**
   * Set a fatal message as the error string.
   *
   * @param[in] format Message format string.
   * @param[in] arguments Argument list for formatted message.
   */
  SCALESTACK_LOCAL
  void _set_fatal_message(const char* format, std::va_list arguments) const;

  /**
   * Log a module message.
   *
   * @param[in] log_level Log level for this message.
   * @param[in] format Message format string.
   * @param[in] arguments Argument list for formatted message.
   */
  SCALESTACK_LOCAL
  void _log_message(log::level log_level,
                    const char* format,
                    std::va_list arguments) const;

  /**
   * Print a help message for this module.
   */
  SCALESTACK_LOCAL
  void _print_help(void);

  /**
   * Print all configuration options for this module.
   */
  SCALESTACK_LOCAL
  void _print_config(void);

  typedef std::vector<module*> modules;
  typedef std::vector<option*> options;

  bool _is_loading;
  bool _is_started;
  bool _need_run;
  bool _need_reconfigure;
  core& _core;
  void* _handle;
  definition* _definition;
  std::string _name;
  std::string _pathname;
  std::string _version;
  std::string _author;
  std::string _title;
  std::string _license;
  std::string _dependencies;
  modules _parents;
  modules::iterator _parent_iterator;
  modules _children;
  modules::iterator _child_iterator;
  options _options;
  options::iterator _option_iterator;

  friend class core;
};

} /* namespace kernel */
} /* namespace scalestack */

/**
 * Macro for creating a module definition structure. This uses information
 * from the plugin.ini file.
 */
#define SCALESTACK_KERNEL_MODULE(options, start, stop, run, reconfigure) \
  SCALESTACK_API scalestack::kernel::module::definition _scalestack_kernel_module_definition = \
  { \
    SCALESTACK_CURRENT_API_VERSION, \
    SCALESTACK_QUOTE_VALUE(PANDORA_MODULE_NAME), \
    PANDORA_MODULE_VERSION, \
    PANDORA_MODULE_AUTHOR, \
    PANDORA_MODULE_TITLE, \
    SCALESTACK_QUOTE_VALUE(PANDORA_MODULE_LICENSE), \
    PANDORA_MODULE_DEPENDENCIES, \
    options, \
    start, \
    stop, \
    run, \
    reconfigure \
  }

#endif /* SCALESTACK_KERNEL_MODULE_H */
