/*
 * Scale Stack
 *
 * Copyright 2010 Eric Day
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * @file
 * @brief Option Declarations
 */

#ifndef SCALESTACK_KERNEL_OPTION_H
#define SCALESTACK_KERNEL_OPTION_H

#include <string>
#include <utility>
#include <vector>

#include <scalestack/kernel/macros.h>

namespace scalestack
{
namespace kernel
{

class module;

/**
 * This class represents a module option and provides various parsing methods
 * for values.
 */
class SCALESTACK_API option
{
public:

  ~option();

  /**
   * Get the module object this option belongs to.
   */
  module& get_module(void) const;

  /**
   * Get the name of the option.
   */
  const std::string& get_name(void) const;

  /**
   * Get the help message of the option.
   */
  const std::string& get_help(void) const;

  /**
   * Get the value name of the option.
   */
  const std::string& get_value_name(void) const;

  /**
   * Get the default value of the option.
   */
  const std::string& get_default_value(void) const;

  /**
   * Get the value of the option.
   */
  const std::string& get_value(void) const;

  /**
   * Get the value of the option converted to a int.
   */
  int get_int_value(void) const;

  /**
   * Get the value of the option converted to a size_t.
   */
  size_t get_size_value(void) const;

  /**
   * Get the value of the option converted to a boolean.
   */
  bool get_bool_value(void) const;

  /**
   * Get the value of the option converted to a list. The format of the list
   * is comma seperated.
   */
  std::vector<std::string> get_list_value(void) const;

  /**
   * Check to see if the associated module asked for this to be a valid
   * option.
   */
  bool is_valid(void) const;

private:

  /**
   * Construct an option instance.
   *
   * @param[in] module Module that created this option.
   * @param[in] name Name of option to create.
   */
  SCALESTACK_LOCAL
  option(module& module, const std::string& name);

  /**
   * Don't allow copying of objects.
   */
  SCALESTACK_LOCAL
  option(const option&);

  /**
   * Don't allow assignment of objects.
   */
  SCALESTACK_LOCAL
  option& operator=(const option&);

  /**
   * Set the help string of the option.
   */
  SCALESTACK_LOCAL
  void _set_help(const std::string& help);

  /**
   * Set the value name of the option.
   */
  SCALESTACK_LOCAL
  void _set_value_name(const std::string& value_name);

  /**
   * Set the default value of the option.
   */
  SCALESTACK_LOCAL
  void _set_default_value(const std::string& default_value);

  /**
   * Get the value of the option.
   */
  SCALESTACK_LOCAL
  void _set_value(const std::string& value, bool overwrite);

  /**
   * Set whether this is a valid option or not.
   */
  SCALESTACK_LOCAL
  void _set_valid(bool valid);

  /**
   * Print a help message for this option.
   */
  SCALESTACK_LOCAL
  void _print_help(void);

  /**
   * Print the configuration line for this option.
   */
  SCALESTACK_LOCAL
  void _print_config(void);

  bool _is_valid;
  bool _is_set;
  module& _module;
  const std::string _name;
  std::string _help;
  std::string _value_name;
  std::string _default_value;
  std::string _value;

  friend class module;
};

} /* namespace kernel */
} /* namespace scalestack */

#endif /* SCALESTACK_KERNEL_OPTION_H */
