/*
 * Scale Stack
 *
 * Copyright 2010 Eric Day
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * @file
 * @brief Kernel Test Declarations and Macro Definitions
 */

#ifndef SCALESTACK_KERNEL_TEST_H
#define SCALESTACK_KERNEL_TEST_H

#include <cassert>
#include <exception>
#include <iostream>
#include <string>

#include <scalestack/kernel/core.h>

using namespace std;
using namespace scalestack;

/**
 * Default event service to use.
 */
#define TEST_EVENT_SERVICE "event::libevent"

/**
 * Begin a test suite block, there should only be one of these per source
 * file.
 */
#define TEST_SUITE_BEGIN \
int main(int _test_argc, char* _test_argv[]) \
{ \
  size_t TEST_LOG_THRESHOLD = 0; \
  if (_test_argc == 2) \
    TEST_LOG_THRESHOLD = string(_test_argv[1]).size(); \
  if (TEST_LOG_THRESHOLD > 0) \
    cout << endl << __FILE__ << endl;

/**
 * Finish the test suite block.
 */
#define TEST_SUITE_END \
  return 0; \
}

/**
 * Begin a test block without defining a kernel core object.
 */
#define TEST_BEGIN_NO_CORE(name) \
if (TEST_LOG_THRESHOLD > 0) \
  cout << "---------- TESTING " << #name << endl; \
{

/**
 * Begin a test block with a default kernel core object named 'core'.
 */
#define TEST_BEGIN(name) \
  TEST_BEGIN_NO_CORE(name) \
  kernel::core core; \
  core.set_threshold(TEST_LOG_THRESHOLD); \
  core.add_module_option_value("path", "", "scalestack/.libs");

/**
 * End a test block.
 */
#define TEST_END \
} \
if (TEST_LOG_THRESHOLD > 0) \
  cout << "---------- PASSED" << endl << endl;

/**
 * Begin an exception test block. This tests to make sure an exception
 * is thrown.
 */
#define TEST_EXCEPTION_BEGIN \
{ \
  bool _test_exception = false; \
  try \
  {

/**
 * End an exception test block.
 */
#define TEST_EXCEPTION_END \
  } \
  catch (exception&) \
  { \
    _test_exception = true; \
  } \
  assert(_test_exception); \
}

/**
 * Test a single line of code to make sure it throws an exception.
 */
#define TEST_EXCEPTION(code) \
do \
  TEST_EXCEPTION_BEGIN \
  code; \
  TEST_EXCEPTION_END \
while (false)

#endif /* SCALESTACK_KERNEL_TEST_H */
