/*
 * Scale Stack
 *
 * Copyright 2010 Eric Day
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * @file
 * @brief Kernel Core Tests
 */

#include "config.h"

#include <cstdlib>
#include <cstring>

#include <scalestack/kernel/service.h>
#include <scalestack/kernel/test.h>

TEST_SUITE_BEGIN

TEST_BEGIN(constructor)
TEST_END

TEST_BEGIN(parse_arguments)
  const char* argv[] =
  {
    "v",
    "-h",
    "--help",
    "test_child"
    "test_child=value1",
    "test_parent.option1",
    "test_parent.option2=value2"
  };
  core.parse_arguments(sizeof(argv) / sizeof(const char*), argv);
TEST_END

TEST_BEGIN(parse_paths)
  const char* argv[] =
  {
    "p=/path/to/modules",
    "path=/another/path/to/modules"
  };
  core.parse_arguments(sizeof(argv) / sizeof(const char*), argv);
TEST_END

TEST_BEGIN(parse_bad_paths)
  const char* argv[] =
  {
    "p",
    "path"
  };
  size_t argc = sizeof(argv) / sizeof(const char*);
  TEST_EXCEPTION(core.parse_arguments(argc, argv));
TEST_END

TEST_BEGIN(parse_files)
  const char* argv[] =
  {
    SCALESTACK_QUOTE(SCALESTACK_SOURCE_PATH)
    "f=scalestack/kernel/test/config_file1",
    SCALESTACK_QUOTE(SCALESTACK_SOURCE_PATH)
    "file=scalestack/kernel/test/config_file2"
  };
  core.parse_arguments(sizeof(argv) / sizeof(const char*), argv);
TEST_END

TEST_BEGIN(bad_option)
  const char* argv[] =
  {
    ""
  };
  size_t argc = sizeof(argv) / sizeof(const char*);
  TEST_EXCEPTION(core.parse_arguments(argc, argv));
  argv[0] = ".option=value";
  TEST_EXCEPTION(core.parse_arguments(argc, argv));
  argv[0] = "=value";
  TEST_EXCEPTION(core.parse_arguments(argc, argv));
TEST_END

TEST_BEGIN_NO_CORE(long_option)
  kernel::core core;
  const size_t long_option_size = 1024 * 128;
  char* long_option = static_cast<char*>(malloc(long_option_size));
  assert(long_option != NULL);
  const char* argv[] =
  {
    long_option
  };

  memset(long_option, 'A', long_option_size - 1);
  long_option[long_option_size - 1] = '\0';
  long_option[long_option_size / 3] = '.';
  long_option[(long_option_size / 3) * 2] = '=';

  core.parse_arguments(sizeof(argv) / sizeof(const char*), argv);
  free(long_option);
TEST_END

TEST_BEGIN_NO_CORE(long_option_list)
  kernel::core core;
  const size_t argc = 1024 * 128;
  const char** argv = static_cast<const char**>(malloc(sizeof(char*) * argc));

  for (size_t x = 0; x < argc; x++)
    argv[x] = const_cast<char*>("module.option=value");

  core.parse_arguments(argc, argv);
  free(argv);
TEST_END

TEST_BEGIN(add_module_option_value)
  core.add_module_option_value("module", "option", "value");
  core.get_module("module");
  TEST_EXCEPTION(core.add_module_option_value("", "option", "value"));
  core.add_module_option_value("module", "", "value");
  core.add_module_option_value("module", "option", "");
  TEST_EXCEPTION(core.add_module_option_value("", "", "value"));
  TEST_EXCEPTION(core.add_module_option_value("", "option", ""));
  core.add_module_option_value("module", "", "");
  TEST_EXCEPTION(core.add_module_option_value("", "", ""));
TEST_END

TEST_BEGIN(set_signal_handlers)
  core.set_signal_handlers();
  TEST_EXCEPTION(core.set_signal_handlers());
TEST_END

TEST_BEGIN(handle_signal)
  core.handle_signal(0);
TEST_END

TEST_BEGIN(get_or_add_module)
  kernel::module& module = core.get_or_add_module("module");
  assert(&core.get_or_add_module("module") == &module);

  kernel::module& module2 = core.get_or_add_module("module2");
  assert(&core.get_or_add_module("module2") == &module2);
  assert(&core.get_or_add_module("module2") != &module);
TEST_END

TEST_BEGIN(get_module)
  kernel::module& module = core.get_or_add_module("module");
  assert(&core.get_module("module") == &module);
  TEST_EXCEPTION(core.get_module("bad_module"));
TEST_END

TEST_BEGIN(get_module_count)
  assert(core.get_module_count() == 0);
  core.get_or_add_module("module");
  assert(core.get_module_count() == 1);
  core.get_or_add_module("module2");
  assert(core.get_module_count() == 2);
  core.get_or_add_module("module2");
  assert(core.get_module_count() == 2);
TEST_END

TEST_BEGIN(get_or_add_module_normalize)
  assert(core.get_module_count() == 0);
  core.get_or_add_module("module_test");
  assert(core.get_module_count() == 1);
  core.get_or_add_module("MODULE_TEST");
  assert(core.get_module_count() == 1);
  core.get_or_add_module("module::test");
  assert(core.get_module_count() == 1);
TEST_END

TEST_BEGIN(module_scan)
  assert(core.begin_module() == NULL);
  assert(core.next_module() == NULL);

  kernel::module& module = core.get_or_add_module("module");
  assert(core.begin_module() == &module);
  assert(core.next_module() == NULL);

  kernel::module& module2 = core.get_or_add_module("module2");
  assert(core.begin_module() == &module);
  assert(core.next_module() == &module2);
  assert(core.next_module() == NULL);
TEST_END

TEST_BEGIN(add_service)
  kernel::module& module = core.get_or_add_module("module");
  kernel::service service(module);;
  core.add_service("service1", &service);
  core.add_service("service2", &service);
  TEST_EXCEPTION(core.add_service("service2", &service));
  assert(&service.get_module() == &module);
TEST_END

TEST_BEGIN(remove_service)
  kernel::module& module = core.get_or_add_module("module");
  kernel::service service(module);
  kernel::service service2(module);
  core.add_service("service1", &service);
  core.add_service("service2", &service2);
  assert(core.remove_service("service1") == &service);
  assert(core.remove_service("service2") == &service2);
  TEST_EXCEPTION(core.remove_service("service2"));
TEST_END

TEST_BEGIN(get_service)
  kernel::module& module = core.get_or_add_module("module");
  kernel::service service(module);
  kernel::service service2(module);
  core.add_service("service1", &service);
  core.add_service("service2", &service2);
  assert(&core.get_service("service1") == &service);
  assert(&core.get_service("service1") == &service);
  assert(&core.get_service("service2") == &service2);
  TEST_EXCEPTION(core.get_service("service3"));
TEST_END

class log_test: public kernel::log
{
public:
  size_t errors;
  log_test(void): errors() {}
  void write_log(level log_level, const char*, const char*)
  {
    if (log_level == kernel::log::LEVEL_ERROR)
      errors++;
  }
};

TEST_BEGIN_NO_CORE(add_log)
  kernel::core core;
  log_test log;
  core.add_log("log1", &log);
  core.add_log("log2", &log);
  TEST_EXCEPTION(core.add_log("log2", &log));
TEST_END

TEST_BEGIN_NO_CORE(remove_log)
  kernel::core core;
  log_test log;
  log_test log2;
  core.add_log("log1", &log);
  core.add_log("log2", &log2);
  assert(core.remove_log("log1") == &log);
  assert(core.remove_log("log2") == &log2);
  TEST_EXCEPTION(core.remove_log("log2"));
TEST_END

TEST_BEGIN_NO_CORE(get_log)
  kernel::core core;
  log_test log;
  log_test log2;
  core.add_log("log1", &log);
  core.add_log("log2", &log2);
  assert(&core.get_log("log1") == &log);
  assert(&core.get_log("log1") == &log);
  assert(&core.get_log("log2") == &log2);
  TEST_EXCEPTION(core.get_log("log3"));
TEST_END

TEST_BEGIN_NO_CORE(unremoved_log)
  cout << _("Testing unremoved logs, ignore the following output:") << endl;
  log_test log1;
  log_test log2;
  /* Put this in a nested scope so it can destruct before end of test. */
  {
    kernel::core core;

    /* Turn on verbose logging so we can catch errors. */
    const char* argv[] =
    {
      "v"
    };
    core.parse_arguments(sizeof(argv) / sizeof(const char*), argv);

    /* The core destructor sends errors when there is more than one unremoved
       log (one for the built-in console log). Add two logs so more than one
       still exists. */
    core.add_log("log1", &log1);
    core.add_log("log2", &log2);

    /* Core goes out of scope here and destructs. */
  }

  /* There should be three errors, one for each existing log plus console. */
  assert(log1.errors == 3);
TEST_END

TEST_BEGIN(run)
  cout << _("Testing exception message, ignore the following output:") << endl;
  bool run_failed = false;
  try
  {
    core.run();
  }
  catch (exception& e)
  {
    cout << e.what() << endl;
    run_failed = true;
  }
  assert(run_failed);
TEST_END

TEST_BEGIN(run_verbose)
  const char* argv[] =
  {
    "vvv",
  };
  core.parse_arguments(sizeof(argv) / sizeof(const char*), argv);
  cout << _("Testing verbose messages, ignore the following output:") << endl;
  TEST_EXCEPTION(core.run());
TEST_END

TEST_BEGIN(shutdown)
  core.shutdown();
TEST_END

TEST_BEGIN(print_help)
  const char* argv[] =
  {
    "help",
  };
  core.parse_arguments(sizeof(argv) / sizeof(const char*), argv);
  cout << _("Testing help messages, ignore the following output:") << endl;
  core.run();
TEST_END

TEST_BEGIN(alias)
  kernel::module& module = core.get_or_add_module("module");
  kernel::module& module2 = core.get_or_add_module("module2");
  core.add_module_option_value("module_a", "", "module");
  assert(&core.get_module("module_a") == &module);
  core.add_module_option_value("module_a", "", "module2");
  assert(&core.get_module("module_a") == &module2);
TEST_END

TEST_SUITE_END
