/*
 * Scale Stack
 *
 * Copyright 2010 Eric Day
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * @file
 * @brief Kernel Module Tests
 */

#include "config.h"

#include <signal.h>
#include <sys/types.h>
#include <unistd.h>

#include <scalestack/kernel/module.h>
#include <scalestack/kernel/test.h>

TEST_SUITE_BEGIN

TEST_BEGIN(constructor)
  kernel::module& module = core.get_or_add_module("module");
  assert(&module.get_core() == &core);
  assert(module.get_name() == "module");
  assert(module.get_pathname() == "");
  assert(module.get_version() == "");
  assert(module.get_author() == "");
  assert(module.get_title() == "");
  assert(module.get_license() == "");
  assert(module.get_dependencies() == "");
  assert(module.get_parent_count() == 0);
  assert(module.get_child_count() == 0);
  assert(module.is_loaded() == false);
TEST_END

TEST_BEGIN(parent_scan)
  kernel::module& module = core.get_or_add_module("module");
  assert(module.begin_parent() == NULL);
  assert(module.next_parent() == NULL);
TEST_END

TEST_BEGIN(child_scan)
  kernel::module& module = core.get_or_add_module("module");
  assert(module.begin_child() == NULL);
  assert(module.next_child() == NULL);
TEST_END

TEST_BEGIN(add_option)
  kernel::module& module = core.get_or_add_module("module");
  module.add_option("option", "Help", "VALUE_NAME", "default_value");
TEST_END

TEST_BEGIN(get_option)
  kernel::module& module = core.get_or_add_module("module");
  kernel::option& option = module.add_option("option",
                                             "Help",
                                             "VALUE_NAME",
                                             "default_value");
  assert(&module.get_option("option") == &option);
  TEST_EXCEPTION(module.get_option("Bad_option"));
TEST_END

TEST_BEGIN(get_option_count)
  kernel::module& module = core.get_or_add_module("module");
  assert(module.get_option_count() == 0);
  module.add_option("option", "Help", "VALUE_NAME", "default_value");
  assert(module.get_option_count() == 1);
  module.add_option("option2", "Help", "VALUE_NAME", "default_value");
  assert(module.get_option_count() == 2);
TEST_END

TEST_BEGIN(option_scan)
  kernel::module& module = core.get_or_add_module("module");
  assert(module.begin_option() == NULL);
  assert(module.next_option() == NULL);

  kernel::option& option = module.add_option("option",
                                             "Help",
                                             "VALUE_NAME",
                                             "default_value");
  assert(module.begin_option() == &option);
  assert(module.next_option() == NULL);

  kernel::option& option2 = module.add_option("option2",
                                              "Help",
                                              "VALUE_NAME",
                                              "default_value");
  assert(module.begin_option() == &option);
  assert(module.next_option() == &option2);
  assert(module.next_option() == NULL);
TEST_END

TEST_BEGIN(load_grand_parent)
  kernel::module& module = core.get_or_add_module("kernel::test::module_grand_parent");
  assert(module.get_title() == "");
  assert(module.is_loaded() == false);
  core.run();
  assert(module.get_title() == "module_grand_parent title");
  assert(module.get_parent_count() == 0);
  assert(module.get_child_count() == 0);
TEST_END

TEST_BEGIN(load_parent)
  kernel::module& module = core.get_or_add_module("kernel::test::module_parent");
  assert(module.get_title() == "");
  assert(module.is_loaded() == false);
  TEST_EXCEPTION(core.get_module("kernel::test::module_child"));
  core.run();

  kernel::module& grand_parent = core.get_module("kernel::test::module_grand_parent");

  assert(grand_parent.get_title() == "module_grand_parent title");
  assert(grand_parent.get_parent_count() == 0);
  assert(grand_parent.get_child_count() == 1);
  assert(grand_parent.begin_child() == &module);
  assert(grand_parent.next_child() == NULL);

  assert(module.get_title() == "module_parent title");
  assert(module.get_parent_count() == 1);
  assert(module.begin_parent() == &grand_parent);
  assert(module.next_parent() == NULL);
  assert(module.get_child_count() == 0);
TEST_END

TEST_BEGIN(load_child)
  kernel::module& module = core.get_or_add_module("kernel::test::module_child");
  assert(module.get_title() == "");
  assert(module.is_loaded() == false);
  core.run();

  kernel::module& parent = core.get_module("kernel::test::module_parent");
  kernel::module& grand_parent = core.get_module("kernel::test::module_grand_parent");

  assert(parent.get_title() == "module_parent title");
  assert(parent.get_parent_count() == 1);
  assert(parent.begin_parent() == &grand_parent);
  assert(parent.next_parent() == NULL);
  assert(parent.get_child_count() == 1);
  assert(parent.begin_child() == &module);
  assert(parent.next_child() == NULL);

  assert(grand_parent.get_title() == "module_grand_parent title");
  assert(grand_parent.get_parent_count() == 0);
  assert(grand_parent.get_child_count() == 1);
  assert(grand_parent.begin_child() == &parent);
  assert(grand_parent.next_child() == NULL);

  assert(module.get_title() == "module_child title");
  assert(module.get_parent_count() == 1);
  assert(module.begin_parent() == &parent);
  assert(module.next_parent() == NULL);
  assert(module.get_child_count() == 0);
TEST_END

TEST_BEGIN(open_error)
  core.get_or_add_module("kernel::test::module_open_error");
  TEST_EXCEPTION(core.run());
TEST_END

TEST_BEGIN(sym_error)
  core.get_or_add_module("kernel::test::module_sym_error");
  TEST_EXCEPTION(core.run());
TEST_END

TEST_BEGIN(load_loop)
  core.get_or_add_module("kernel::test::module_loop");
  TEST_EXCEPTION(core.run());
TEST_END

TEST_BEGIN(module_scan)
  kernel::module& module = core.get_or_add_module("kernel::test::module_grand_parent");
  assert(core.begin_module() == &module);
  assert(core.next_module() == NULL);

  core.run();

  kernel::module& module2 = core.get_or_add_module("kernel::test::module_parent");

  assert(core.begin_module() == &module);
  assert(core.next_module() == &module2);
  assert(core.next_module() == NULL);
TEST_END

TEST_BEGIN_NO_CORE(verbose)
  cout << _("Testing verbose messages, ignore the following output:") << endl;
  /* Put this in a nested scope so it can destruct before end of test. */
  {
    kernel::core core;
    core.add_module_option_value("path", "", "scalestack/.libs");
    const char* argv[] =
    {
      "vvv"
    };
    core.parse_arguments(sizeof(argv) / sizeof(const char*), argv);
    core.get_or_add_module("kernel::test::module_grand_parent");
    core.run();
  }
TEST_END

TEST_BEGIN(signal_shutdown)
  core.get_or_add_module("kernel::test::module_grand_parent");
  core.set_signal_handlers();
  kill(getpid(), SIGQUIT);
  core.run();
TEST_END

TEST_BEGIN(print_help)
  core.get_or_add_module("kernel::test::option_basic");
  core.get_or_add_module("kernel::test::module_grand_parent");
  const char* argv[] =
  {
    "help",
  };
  core.parse_arguments(sizeof(argv) / sizeof(const char*), argv);
  cout << _("Testing help messages, ignore the following output:") << endl;
  core.run();
TEST_END

TEST_BEGIN(check_invalid_options)
  core.get_or_add_module("kernel::test::option_basic");
  core.add_module_option_value("kernel::test::option_basic", "Invalid", "value");
  TEST_EXCEPTION(core.run());
TEST_END

TEST_SUITE_END
