/*
 * Scale Stack
 *
 * Copyright 2010 Chris Behrens
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * @file
 * @brief Network Datagram Declarations
 */

#ifndef SCALESTACK_NETWORK_DATAGRAM_H
#define SCALESTACK_NETWORK_DATAGRAM_H

#include <stdint.h>
#include <sys/socket.h>
#include <sys/types.h>

#include <scalestack/event/handler.h>
#include <scalestack/kernel/macros.h>
#include <scalestack/network/datagram_service.h>

namespace scalestack
{
namespace kernel
{

class module;

} /* namespace kernel */

namespace network
{

class datagram_provider;

/**
 * This is the base class for modules using datagrams.
 */
class SCALESTACK_API datagram
{
public:

  datagram(datagram_service& datagram_service);

  virtual ~datagram();

  /**
   * This is called to notify derived classes that the datagram is ready
   * for service. The default behavior is to ignore.
   */
  virtual void ready(void);

  /**
   * This is called when data has been received and needs to be handled by
   * derived classes.
   *
   * @param[in] buffer Data that was read.
   * @param[in] size Size of data available in buffer.
   * @param[in] peer Socket address the packet was received from.
   * @param[in] peer_size Socket address length.
   * @return Amount of data that was consumed.
   */
  virtual size_t receive(uint8_t* buffer,
                         size_t size,
                         struct sockaddr& peer,
                         socklen_t peer_size) = 0;

  /**
   * Notify the connection that the receive buffer has been consumed.
   * This is required after a receive() call was not able to consume
   * all the data. This can be called at a later time to remove the data
   * from the read buffer, such as in flush_send(). After the receive buffer
   * has been flushed, the connection will start watching for new data.
   *
   * @param[in] size The amount of data to consume.
   */
  void consume(size_t size);

  /**
   * Send data to the connection. This is not buffered. If data could not
   * be written, flush_send() will be called when it can be.
   *
   * @param[in] buffer Buffer to write.
   * @param[in] size Size of buffer.
   * @param[in] peer Socket address to send to.
   * @param[in] peer_size Socket address length.
   * @return Amount of data sent.
   */
  size_t send(uint8_t* buffer,
              size_t size,
              const struct sockaddr& peer,
              socklen_t peer_size);

  /**
   * Same as send(), but use the default peer address this datagram was
   * created with.
   *
   * @param[in] buffer Buffer to write.
   * @param[in] size Size of buffer.
   * @return Amount of data sent.
   */
  size_t send(uint8_t* buffer, size_t size);

  /**
   * This is called to notify derived classes that they should trying sending
   * data again by calling send(). This is called when the connection is
   * ready for writing after a call to send() could not write data.
   */
  virtual void flush_send(void) = 0;

  /*
   * event::Handler methods to pass through.
   */

  /**
   * See event::handler::remove().
   */
  void remove(void);

  /**
   * See event::handler::error().
   */
  virtual void error(void);

  /**
   * See event::handler::set_timer().
   */
  void set_timer(time_t seconds);

  /**
   * See event::handler::timer_expired().
   */
  virtual void timer_expired(void);

  /**
   * See event::handler::run_now().
   */
  void run_now(void);

  /**
   * See event::handler::run().
   */
  virtual void run(void);

protected:

  kernel::module& _module;
  datagram_service& _datagram_service;

private:

  /**
   * Don't allow copying of objects.
   */
  SCALESTACK_LOCAL
  datagram(const datagram&);

  /**
   * Don't allow assignment of objects.
   */
  SCALESTACK_LOCAL
  datagram& operator=(const datagram&);

  datagram_provider* _datagram_provider;

  friend class datagram_provider;
};

/**
 * This is the base class for classes that provide datagrams.
 */
class SCALESTACK_API datagram_provider: public event::handler
{
public:

  datagram_provider(event::handler_service& handler_service);

  virtual ~datagram_provider();

  /**
   * Get the datagram bound to this provider.
   */
  datagram* get_datagram(void);

  /**
   * Set the datagram for this datagram_provider. This is used while
   * initializing the datagram handler.
   *
   * @param[in] datagram Datagram to bind to this provider.
   */
  void set_datagram(datagram* datagram);

  /**
   * Set the file descriptor to use for this connection.
   *
   * @param[in] file_descriptor File descriptor to set.
   */
  virtual void set_file_descriptor(int file_descriptor);

  /**
   * Create the socket using a random local address.
   */
  void create(int family);

  /**
   * Create the socket and bind to a specific local address.
   *
   * @param[in] local Local address to bind to.
   * @param[in] local_size Size of the address structure.
   */
  void create(const struct sockaddr& local, socklen_t local_size);

  /**
   * Set the initial peer address to send packets to.
   *
   * @param[in] peer Peer address to set.
   * @param[in] peer_size Size of peer address structure.
   */
  void set_peer(const struct sockaddr& peer, socklen_t peer_size);

  /**
   * See datagram::ready().
   */
  void ready(void);

  /**
   * See datagram::receive().
   */
  size_t receive(uint8_t* buffer,
                 size_t size,
                 struct sockaddr& peer,
                 socklen_t peer_size);

  /**
   * See datagram::consume().
   */
  void consume(size_t size);

  /**
   * See datagram::send().
   */
  size_t send(uint8_t* buffer,
              size_t size,
              const struct sockaddr& peer,
              socklen_t peer_size);

  /**
   * See datagram::send().
   */
  size_t send(uint8_t* buffer, size_t size);

  /**
   * See datagram::flush_send().
   */
  void flush_send(void);

  /*
   * event::Handler methods to pass through.
   */

  /**
   * See event::handler::error().
   */
  void error(void);

  /**
   * See event::handler::timer_expired().
   */
  void timer_expired(void);

  /**
   * See event::handler::run().
   */
  void run(void);

private:

  /**
   * Don't allow copying of objects.
   */
  SCALESTACK_LOCAL
  datagram_provider(const datagram_provider&);

  /**
   * Don't allow assignment of objects.
   */
  SCALESTACK_LOCAL
  datagram_provider& operator=(const datagram_provider&);

  /**
   * This will receive data from the connection into the buffer and call
   * receive(). This implements a method in the base class so does not have
   * the _ prefix.
   *
   * @param[in] file_descriptor File descriptor to read data on.
   */
  void read_ready(int file_descriptor);

  /**
   * This just calls flush_send() since we don't buffer any data. This
   * implements a method in the base class so does not have the _ prefix.
   *
   * @param[in] file_descriptor File descriptor to write data to.
   */
  void write_ready(int file_descriptor);

  /**
   * Perform the receive buffer consumption, used by both read_ready()
   * and consume().
   *
   * @param[in] size Amount of buffer to consume.
   */
  SCALESTACK_LOCAL
  void _consume(size_t size);

  static const size_t _receive_buffer_max_size = 64 * 1024;

  int _file_descriptor;
  socklen_t _peer_size;
  size_t _receive_buffer_size;
  datagram* _datagram;
  struct sockaddr* _peer;
  struct sockaddr_storage _peer_storage;
  uint8_t _receive_buffer[_receive_buffer_max_size];
};

/*
 * Public datagram methods.
 */

inline datagram::datagram(datagram_service& datagram_service):
  _module(datagram_service.get_module()),
  _datagram_service(datagram_service),
  _datagram_provider(NULL)
{
}

inline datagram::~datagram()
{
}

inline void datagram::ready(void)
{
}

inline void datagram::consume(size_t size)
{
  _datagram_provider->consume(size);
}

inline size_t datagram::send(uint8_t* buffer,
                             size_t size,
                             const struct sockaddr& peer,
                             socklen_t peer_size)
{
  return _datagram_provider->send(buffer, size, peer, peer_size);
}

inline size_t datagram::send(uint8_t* buffer, size_t size)
{
  return _datagram_provider->send(buffer, size);
}

inline void datagram::remove(void)
{
  _datagram_provider->remove();
}

inline void datagram::error(void)
{
  _datagram_provider->remove();
}

inline void datagram::set_timer(time_t seconds)
{
  _datagram_provider->set_timer(seconds);
}

inline void datagram::timer_expired(void)
{
}

inline void datagram::run_now(void)
{
  _datagram_provider->run_now();
}

inline void datagram::run(void)
{
}

/*
 * Public datagram_provider methods.
 */

inline datagram* datagram_provider::get_datagram(void)
{
  return _datagram;
}

inline void datagram_provider::set_datagram(datagram* datagram)
{
  _datagram = datagram;
  _datagram->_datagram_provider = this;
}

inline void datagram_provider::ready(void)
{
  _datagram->ready();
}

inline size_t datagram_provider::receive(uint8_t* buffer,
                                         size_t size,
                                         struct sockaddr& peer,
                                         socklen_t peer_size)
{
  return _datagram->receive(buffer, size, peer, peer_size);
}

inline void datagram_provider::flush_send(void)
{
  _datagram->flush_send();
}

inline void datagram_provider::error(void)
{
  _datagram->error();
}

inline void datagram_provider::timer_expired(void)
{
  _datagram->timer_expired();
}

} /* namespace network */
} /* namespace scalestack */

#endif /* SCALESTACK_NETWORK_DATAGRAM_H */
