/*
 * Scale Stack
 *
 * Copyright 2010 Eric Day
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * @file
 * @brief IP Address Service Definitions
 */

#include "config.h"

#include <cstring>

#include <scalestack/kernel/module.h>
#include <scalestack/network/ip/address_service.h>

using namespace std;

namespace scalestack
{
namespace network
{
namespace ip
{

/*
 * Public methods.
 */

address_service::address_service(kernel::module& module):
  _module(module),
  _local_only(),
  _address_info_list(NULL),
  _addresses(),
  _address(_addresses.end()),
  _address_options()
{
}

address_service::~address_service()
{
  if (_address_info_list != NULL)
    freeaddrinfo(_address_info_list);
}

void address_service::parse_addresses(const std::string& addresses,
                                      bool local_only,
                                      const struct addrinfo& address_options)
{
  /* Parse list by commas first. */
  vector<string> list;
  string::size_type current = 0;

  do
  {
    string::size_type next = addresses.find_first_of(",", current);
    list.push_back(addresses.substr(current, next - current));
    current = addresses.find_first_not_of(",", next);
  }
  while (current != string::npos);

  /* Parse addresses. */
  for (vector<string>::const_iterator address = list.begin();
       address != list.end();
       ++address)
  {
    string::size_type port = address->find_last_of(":");
    if (port == string::npos)
      _addresses.push_back(make_pair(*address, ""));
    else
    {
      _addresses.push_back(make_pair(address->substr(0, port),
                                     address->substr(port + 1)));
    }
  }

  memcpy(&_address_options, &address_options, sizeof(struct addrinfo));
  _local_only = local_only;
  reset();
}

const struct addrinfo* address_service::get_next(void)
{
  if (_address == _addresses.end())
    return NULL;

  if (_address_info_list != NULL)
  {
    freeaddrinfo(_address_info_list);
    _address_info_list = NULL;
  }

  int return_code = ::getaddrinfo(_address->first == "" ? NULL : _address->first.c_str(),
                                  _address->second == "" ? NULL : _address->second.c_str(),
                                   &_address_options,
                                   &_address_info_list);
  if (return_code != 0)
    _module.log_fatal(_("Failed getaddrinfo: %s"), gai_strerror(return_code));

  ++_address;

  return _address_info_list;
}

void address_service::get_all(void)
{
  reset();

  while (get_next() != NULL)
  {
    if (_local_only)
      add_address_list(_address_info_list, NULL);
    else
      add_address_list(NULL, _address_info_list);
  }
}

void address_service::reset(void)
{
  _address = _addresses.begin();

  if (_address_info_list != NULL)
  {
    freeaddrinfo(_address_info_list);
    _address_info_list = NULL;
  }
}

} /* namespace ip */
} /* namespace network */
} /* namespace scalestack */
