/*
 * Scale Stack
 *
 * Copyright 2010 Eric Day
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * @file
 * @brief IP Address Service Declarations
 */

#ifndef SCALESTACK_NETWORK_IP_ADDRESS_SERVICE_H
#define SCALESTACK_NETWORK_IP_ADDRESS_SERVICE_H

#include <string>
#include <netdb.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <vector>

#include <scalestack/kernel/macros.h>

namespace scalestack
{
namespace kernel
{

class module;

} /* namespace kernel */

namespace network
{
namespace ip
{

/**
 * This class parses IP address lists for other protocol modules like TCP
 * and UDP.
 */
class SCALESTACK_API address_service
{
public:

  address_service(kernel::module& module);

  virtual ~address_service();

  /**
   * Parse a string containing a list of addresses. If local_only is true,
   * the list will be a simple [HOST][:PORT]... pair. For example, the string:
   *
   * 10.0.0.1:12345,:12346
   *
   * Would parse as two different local addresses and the peer argument
   * in add_address_list() will always be NULL. The first would be for
   * 10.0.0.1:12345, and the second would be on all local addresses
   * (INADDR_ANY) with the pot 12346.
   *
   * If local_only is false, both local and peer addresses will be parsed. For
   * example, the string:
   *
   * 10.0.0.1+10.0.0.2:12345,10.0.0.3:12345
   *
   * Would parse as two different addresses. The first would have a local
   * address of 10.0.0.1 with a random local port and a peer address:port
   * of 10.0.0.2:12345. The second would have a random local address:port
   * and a peer address:port of 10.0.0.3:12345.
   *
   * @param[in] addresses List of addresses to parse.
   * @param[in] local_only Whether to only parse local addresses or not. This
   *  should be true when parsing addresses then will be used for listening.
   * @param[in] address_options Options to use when parsing addresses.
   */
  void parse_addresses(const std::string& addresses,
                       bool local_only,
                       const struct addrinfo& address_options);

  const struct addrinfo* get_next(void);
  void get_all(void);
  void reset(void);

  /**
   * Method that gets called for each address parse_addresses is able
   * to parse.
   *
   * @param[in] local Local address that was parsed.
   * @param[in] peer Peer address that was parsed. This will always be NULL
   *  when the local_only argument is true in parse_addresses().
   */
  virtual void add_address_list(const struct addrinfo* local,
                                const struct addrinfo* peer) = 0;

protected:

  kernel::module& _module;

private:

  /**
   * Don't allow copying of objects.
   */
  SCALESTACK_LOCAL
  address_service(const address_service&);

  /**
   * Don't allow assignment of objects.
   */
  SCALESTACK_LOCAL
  address_service& operator=(const address_service&);

  bool _local_only;
  struct addrinfo* _address_info_list;
  std::vector<std::pair<std::string,std::string> > _addresses;
  std::vector<std::pair<std::string,std::string> >::const_iterator _address;
  struct addrinfo _address_options;
};

} /* namespace ip */
} /* namespace network */
} /* namespace scalestack */

#endif /* SCALESTACK_NETWORK_IP_ADDRESS_SERVICE_H */
