/*
 * Scale Stack
 *
 * Copyright 2010 Eric Day
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * @file
 * @brief Network IP TCP Connection Definitions
 */

#include "config.h"

#include <cerrno>
#include <cstring>
#include <sys/types.h> /* Must be before socket headers for Free BSD. */
#include <netdb.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <sys/socket.h>
#include <unistd.h>

#include <scalestack/kernel/module.h>
#include <scalestack/network/ip/tcp/connection.h>
#include <scalestack/network/ip/tcp/connection_service.h>

using namespace std;

namespace scalestack
{
namespace network
{
namespace ip
{
namespace tcp
{

/*
 * Public methods.
 */

connection::connection(connection_service& connection_service):
  network::stream_provider(connection_service)
{
}

connection::~connection()
{
  if (_address_list != NULL)
    freeaddrinfo(_address_list);
}

void connection::set_file_descriptor(int file_descriptor)
{
  network::stream_provider::set_file_descriptor(file_descriptor);

  if (file_descriptor != -1)
  {
    int return_code = 1;
    return_code = setsockopt(file_descriptor, IPPROTO_TCP, TCP_NODELAY,
                             &return_code, static_cast<socklen_t>(sizeof(int)));
    if (return_code == -1 && errno != EOPNOTSUPP)
    {
      _module.log_error(_("Could not set TCP_NODELAY: %s"), strerror(errno));
      error();
      return;
    }
  }
}

void connection::connect(const std::string& address)
{
  std::string host;
  std::string port;

  string::size_type port_offset = address.find_first_of(":");
  if (port_offset == string::npos)
    host = address;
  else
  {
    host = address.substr(0, port_offset);
    port = address.substr(port_offset + 1);
  }

  struct addrinfo addrinfo_options;

  if (_address_list != NULL)
  {
    freeaddrinfo(_address_list);
    _address_list = NULL;
  }

  memset(&addrinfo_options, 0, sizeof(struct addrinfo));
  addrinfo_options.ai_socktype = SOCK_STREAM;
  addrinfo_options.ai_protocol = IPPROTO_TCP;

  int return_code = getaddrinfo(host.c_str(), port.c_str(), &addrinfo_options,
                                &_address_list);
  if (return_code != 0)
  {
    _module.log_error(_("Failed getaddrinfo: %s"), gai_strerror(return_code));
    error();
    return;
  }

  try
  {
    network::stream_provider::connect();
  }
  catch(exception&)
  {
    error();
    return;
  }
}

} /* namespace tcp */
} /* namespace ip */
} /* namespace network */
} /* namespace scalestack */
