/*
 * Scale Stack
 *
 * Copyright 2010 Eric Day
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * @file
 * @brief Network IP TCP Connection Service Definitions
 */

#include "config.h"

#include <cstring>
#include <exception>
#include <netdb.h>
#include <sys/socket.h>

#include <scalestack/kernel/module.h>
#include <scalestack/kernel/option.h>
#include <scalestack/network/stream_service.h>
#include <scalestack/network/ip/tcp/connection.h>
#include <scalestack/network/ip/tcp/connection_service.h>

using namespace std;

namespace scalestack
{
namespace network
{
namespace ip
{
namespace tcp
{

/*
 * Public methods.
 */

void connection_service::options(kernel::module& module, const char* addresses)
{
  module.add_option("addresses",
                    _("List of host:port pairs."),
                    "[HOST][:PORT][,[HOST][:PORT]]...",
                    addresses);
}

connection_service::connection_service(kernel::module& module,
                                       stream_service& stream_service):
  event::handler_service(module),
  _stream_service(stream_service)
{
}

connection_service::~connection_service()
{
  shutdown();
}

network::stream* connection_service::add_accepted(int file_descriptor,
                                                  struct sockaddr& peer,
                                                  socklen_t peer_size)
{
  char host_name[NI_MAXHOST];
  char port_name[NI_MAXSERV];

  int return_code = getnameinfo(&peer,
                                peer_size,
                                host_name, NI_MAXHOST,
                                port_name, NI_MAXSERV,
                                NI_NUMERICHOST | NI_NUMERICSERV);
  if (return_code != 0)
  { 
    _module.log_error(_("Error accepting TCP connection, getnameinfo: %s"),
                     gai_strerror(return_code));
  }
  else
    _module.log_notice(_("Accepted %s:%s"), host_name, port_name);

  connection* connection;
  stream* stream = _add_connection(&connection);

  connection->set_file_descriptor(file_descriptor);

  return stream;
}

network::stream* connection_service::add_connection(const std::string& address)
{
  connection* connection;
  network::stream* stream = _add_connection(&connection);

  connection->connect(address);

  return stream;
}

void connection_service::add_connections(void)
{
#if 0
  struct addrinfo address_options;

  memset(&address_options, 0, sizeof(struct addrinfo));
  address_options.ai_flags  = AI_PASSIVE;
  address_options.ai_family = AF_UNSPEC;
  address_options.ai_socktype = SOCK_STREAM;
  address_options.ai_protocol = IPPROTO_TCP;

  parse_addresses(_module.get_option("addresses").get_value(),
                  true,
                  address_options);
#endif
}

/*
 * Private methods.
 */

void connection_service::remove_handler(event::handler* handler)
{
  _stream_service.remove_stream(static_cast<connection*>(handler)->get_stream());
  delete handler;
}

network::stream* connection_service::_add_connection(connection** connection)
{
  network::stream* stream = NULL;
  *connection = NULL;

  try
  {
    stream = _stream_service.add_stream();
    *connection = new tcp::connection(*this);
    (*connection)->set_stream(stream);
    add_handler(*connection);
  }
  catch (exception&)
  {
    if (*connection != NULL)
      _stream_service.remove_stream(stream);

    throw;
  }

  return stream;
}

#if 0
void connection_service::add_address_list(const struct addrinfo* local,
                                          const struct addrinfo* peer)
{
#if 0
  if (local == NULL)
  {
    _add_socket(NULL, peer);
    return;
  }

  /* We only require one local address from the list to succeed. */
  bool added = false;
  for (; local != NULL; local = local->ai_next)
  {
    try
    {
      _add_socket(local, peer);
      added = true;
    }
    catch (exception&)
    {
    }
  }

  if (!added)
    _module.log_fatal(_("Could not add any TCP connections"));
#endif
(void) local;
(void) peer;
}
#endif

} /* namespace tcp */
} /* namespace ip */
} /* namespace network */
} /* namespace scalestack */
