/*
 * Scale Stack
 *
 * Copyright 2010 Eric Day
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * @file
 * @brief Network IP TCP Connection Service Declarations
 */

#ifndef SCALESTACK_NETWORK_IP_TCP_CONNECTION_SERVICE_H
#define SCALESTACK_NETWORK_IP_TCP_CONNECTION_SERVICE_H

#include <string>
#include <sys/socket.h>
#include <sys/types.h>

#include <scalestack/event/handler_service.h>
#include <scalestack/kernel/macros.h>
#include <scalestack/network/ip/address_service.h>

namespace scalestack
{
namespace network
{

class stream;
class stream_service;

namespace ip
{
namespace tcp
{

class connection;

/**
 * This class manages a set of TCP connections.
 */
class SCALESTACK_API connection_service: public event::handler_service
{
public:

  /**
   * Add common options for TCP connections.
   *
   * @param[in] module Module that the service is associated with.
   * @param[in] addresses Default address list to use.
   */
  static void options(kernel::module& module, const char* addresses);

  connection_service(kernel::module& module, stream_service& stream_service);

  virtual ~connection_service();

  /**
   * See listener_service::add_accepted().
   */
  network::stream* add_accepted(int file_descriptor,
                                struct sockaddr& peer,
                                socklen_t peer_size);

  network::stream* add_connection(void);

  network::stream* add_connection(const std::string& address);

  /**
   * Add connections specified by the 'addresses' variable in the module
   * options and connect to those addresses.
   */
  void add_connections(void);

  /**
   * Add connections specified by the given 'addresses' parameter.
   */
  void add_connections(const std::string& addresses);

private:

  /**
   * Don't allow copying of objects.
   */
  SCALESTACK_LOCAL
  connection_service(const connection_service&);

  /**
   * Don't allow assignment of objects.
   */
  SCALESTACK_LOCAL
  connection_service& operator=(const connection_service&);

  /**
   * See event::handler_service::remove_handler().
   */
  void remove_handler(event::handler* handler);

  /**
   * Add a new connection to use for accept() or connect().
   */
  SCALESTACK_LOCAL
  network::stream* _add_connection(connection** connection);

  stream_service& _stream_service;
};

} /* namespace tcp */
} /* namespace ip */
} /* namespace network */
} /* namespace scalestack */

#endif /* SCALESTACK_NETWORK_IP_TCP_CONNECTION_SERVICE_H */
