/*
 * Scale Stack
 *
 * Copyright 2010 Eric Day
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * @file
 * @brief Network IP TCP Listener Service Definitions
 */

#include "config.h"

#include <cstring>
#include <exception>
#include <netdb.h>
#include <sys/socket.h>

#include <scalestack/kernel/module.h>
#include <scalestack/kernel/option.h>
#include <scalestack/network/ip/tcp/connection_service.h>
#include <scalestack/network/ip/tcp/listener.h>
#include <scalestack/network/ip/tcp/listener_service.h>

using namespace std;

namespace scalestack
{
namespace network
{
namespace ip
{
namespace tcp
{

/*
 * Public methods.
 */

void listener_service::options(kernel::module& module,
                               const std::string& addresses)
{
  module.add_option("addresses",
                    _("List of host:port pairs."),
                    "[HOST][:PORT][,[HOST][:PORT]]...",
                    addresses);
  module.add_option("backlog",
                    _("Backlog value to use for listen()."),
                    "BACKLOG",
                    "64");
}

listener_service::listener_service(kernel::module& module,
                                   connection_service& connection_service):
  network::listener_service(module),
  _connection_service(connection_service)
{
}

listener_service::~listener_service()
{
  shutdown();
}

network::stream* listener_service::add_accepted(int file_descriptor,
                                                struct sockaddr& peer,
                                                socklen_t peer_size)
{
  return _connection_service.add_accepted(file_descriptor, peer, peer_size);
}

void listener_service::add_listeners(void)
{
  add_listeners(_module.get_option("addresses").get_value());
}

void listener_service::add_listeners(const std::string& addresses)
{
  struct addrinfo address_options;

  memset(&address_options, 0, sizeof(struct addrinfo));
  address_options.ai_flags  = AI_PASSIVE;
  address_options.ai_family = AF_UNSPEC;
  address_options.ai_socktype = SOCK_STREAM;
  address_options.ai_protocol = IPPROTO_TCP;

  address_service address_service(_module, *this);
  address_service.parse_addresses(addresses, true, address_options);
  address_service.get_all();
}

/*
 * Private methods.
 */

void listener_service::remove_handler(event::handler* handler)
{
  delete handler;
}

/*
 * Public methods for address_service.
 */

address_service::address_service(kernel::module& module,
                                 listener_service& listener_service):
  ip::address_service(module),
  _listener_service(listener_service)
{
}

void address_service::add_address_list(const struct addrinfo* local,
                                       const struct addrinfo*)
{
  /* We only require one local address from the list to succeed. */
  bool added = false;
  for (; local != NULL; local = local->ai_next)
  {
    try
    {
      new listener(_listener_service, *local);
      added = true;
    }
    catch (exception&)
    {
    }
  }

  if (!added)
    _module.log_fatal(_("Could not add any TCP listeners"));
}

} /* namespace tcp */
} /* namespace ip */
} /* namespace network */
} /* namespace scalestack */
