/*
 * Scale Stack
 *
 * Copyright 2010 Eric Day
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * @file
 * @brief Network IP TCP Listener Service Declarations
 */

#ifndef SCALESTACK_NETWORK_IP_TCP_LISTENER_SERVICE_H
#define SCALESTACK_NETWORK_IP_TCP_LISTENER_SERVICE_H

#include <string>

#include <scalestack/kernel/macros.h>
#include <scalestack/network/ip/address_service.h>
#include <scalestack/network/listener_service.h>

namespace scalestack
{
namespace network
{
namespace ip
{
namespace tcp
{

class listener;
class connection_service;

/**
 * This class manages a set of TCP listener sockets.
 */
class SCALESTACK_API listener_service: public network::listener_service
{
public:

  /**
   * Add common options for Connection listen sockets.
   *
   * @param[in] module Module that the service is associated with.
   * @param[in] addresses Default addresses to listen on.
   */
  static void options(kernel::module& module, const std::string& addresses);

  listener_service(kernel::module& module,
                   connection_service& connection_service);

  ~listener_service();

  /**
   * See network::listener_service::add_accepted().
   */
  network::stream* add_accepted(int file_descriptor,
                                struct sockaddr& peer,
                                socklen_t peer_size);

  /**
   * Add listeners for all addresses specified by the 'addresses' variable
   * in the module options.
   */
  void add_listeners(void);

  /**
   * Add listeners for the list of addesses given.
   */
  void add_listeners(const std::string& addresses);

private:

  /**
   * Don't allow copying of objects.
   */
  SCALESTACK_LOCAL
  listener_service(const listener_service&);

  /**
   * Don't allow assignment of objects.
   */
  SCALESTACK_LOCAL
  listener_service& operator=(const listener_service&);

  /**
   * See event::handler_service::remove_handler().
   */
  SCALESTACK_LOCAL
  void remove_handler(event::handler* handler);

  connection_service& _connection_service;
};

/**
 * This class is used to parse addresses and add them to a listener service.
 */
class address_service: public ip::address_service
{
public:

  address_service(kernel::module& module, listener_service& listener_service);

  /**
   * See ip::address_service::add_address_list().
   */
  void add_address_list(const struct addrinfo* local,
                        const struct addrinfo* peer);

private:
  listener_service& _listener_service;
};

} /* namespace tcp */
} /* namespace ip */
} /* namespace network */
} /* namespace scalestack */

#endif /* SCALESTACK_NETWORK_IP_TCP_LISTENER_SERVICE_H */
