/*
 * Scale Stack
 *
 * Copyright 2010 Chris Behrens
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * @file
 * @brief Network IP UDP Socket Service Definitions
 */

#include "config.h"

#include <cstring>
#include <exception>
#include <netdb.h>
#include <sys/socket.h>

#include <scalestack/kernel/module.h>
#include <scalestack/kernel/option.h>
#include <scalestack/network/datagram_service.h>
#include <scalestack/network/ip/udp/socket.h>
#include <scalestack/network/ip/udp/socket_service.h>

using namespace std;

namespace scalestack
{
namespace network
{
namespace ip
{
namespace udp
{

/*
 * Public methods.
 */

void socket_service::options(kernel::module& module,
                             const std::string& addresses)
{
  module.add_option("addresses",
                    _("List of host:port pairs."),
                    "[HOST][:PORT][,[HOST][:PORT]]...",
                    addresses);
}

socket_service::socket_service(kernel::module& module,
                               datagram_service& datagram_service):
  event::handler_service(module),
  _datagram_service(datagram_service)
{
}

socket_service::~socket_service()
{
  shutdown();
}

network::datagram* socket_service::add_socket(const struct addrinfo *local,
                                              const struct addrinfo *peer)
{
  char host_name[NI_MAXHOST];
  char port_name[NI_MAXSERV];
  int return_code;

  if (local != NULL)
  {
    return_code = getnameinfo(local->ai_addr,
                              local->ai_addrlen,
                              host_name, NI_MAXHOST,
                              port_name, NI_MAXSERV,
                              NI_NUMERICHOST | NI_NUMERICSERV);
    if (return_code == 0)
      _module.log_info(_("UDP socket bound to %s:%s"), host_name, port_name);
    else
      _module.log_info(_("Failed getnameinfo: %s"), gai_strerror(return_code));
  }

  if (peer != NULL)
  {
    return_code = getnameinfo(peer->ai_addr,
                              peer->ai_addrlen,
                              host_name, NI_MAXHOST,
                              port_name, NI_MAXSERV,
                              NI_NUMERICHOST | NI_NUMERICSERV);
    if (return_code == 0)
      _module.log_info(_("UDP socket with peer %s:%s"), host_name, port_name);
    else
      _module.log_info(_("Failed getnameinfo: %s"), gai_strerror(return_code));
  }

  datagram* datagram = NULL;
  socket* socket = NULL;

  try
  {
    datagram = _datagram_service.add_datagram();
    socket = new udp::socket(*this);
    socket->set_datagram(datagram);
    add_handler(socket);
  }
  catch (exception&)
  {
    if (datagram != NULL)
      _datagram_service.remove_datagram(datagram);

    if (socket != NULL)
      delete socket;

    throw;
  }

  if (local == NULL)
  {
    socket->create(peer->ai_family);
    socket->set_peer(*(peer->ai_addr), peer->ai_addrlen);
  }
  else
    socket->create(*(local->ai_addr), local->ai_addrlen);

  socket->ready();

  return datagram;
}

void socket_service::add_sockets(void)
{
  add_sockets(_module.get_option("addresses").get_value());
}

void socket_service::add_sockets(const std::string& addresses)
{
  struct addrinfo address_options;

  memset(&address_options, 0, sizeof(struct addrinfo));
  address_options.ai_flags  = AI_PASSIVE;
  address_options.ai_family = AF_UNSPEC;
  address_options.ai_socktype = SOCK_DGRAM;
  address_options.ai_protocol = IPPROTO_UDP;

  address_service address_service(_module, *this);
  address_service.parse_addresses(addresses, true, address_options);
  address_service.get_all();
}

void socket_service::add_peer_sockets(void)
{
  add_peer_sockets(_module.get_option("addresses").get_value());
}

void socket_service::add_peer_sockets(const std::string& addresses)
{
  struct addrinfo address_options;

  memset(&address_options, 0, sizeof(struct addrinfo));
  address_options.ai_flags  = AI_PASSIVE;
  address_options.ai_family = AF_UNSPEC;
  address_options.ai_socktype = SOCK_DGRAM;
  address_options.ai_protocol = IPPROTO_UDP;

  address_service address_service(_module, *this);
  address_service.parse_addresses(addresses, false, address_options);
  address_service.get_all();
}

/*
 * Private methods.
 */

void socket_service::remove_handler(event::handler* handler)
{
  _datagram_service.remove_datagram(static_cast<socket*>(handler)->get_datagram());
  delete handler;
}

/*
 * Public methods for address_service.
 */

address_service::address_service(kernel::module& module,
                                 socket_service& socket_service):
  ip::address_service(module),
  _socket_service(socket_service)
{
}

void address_service::add_address_list(const struct addrinfo* local,
                                       const struct addrinfo* peer)
{
  if (local == NULL)
  {
    _socket_service.add_socket(NULL, peer);
    return;
  }

  /* We only require one local address from the list to succeed. */
  bool added = false;
  for (; local != NULL; local = local->ai_next)
  {
    try
    {
      _socket_service.add_socket(local, peer);
      added = true;
    }
    catch (exception&)
    {
    }
  }

  if (!added)
    _module.log_fatal(_("Could not add any UDP sockets"));
}

} /* namespace udp */
} /* namespace ip */
} /* namespace network */
} /* namespace scalestack */
