/*
 * Scale Stack
 *
 * Copyright 2010 Chris Behrens
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * @file
 * @brief Network IP UDP Socket Service Declarations
 */

#ifndef SCALESTACK_NETWORK_IP_UDP_SOCKET_SERVICE_H
#define SCALESTACK_NETWORK_IP_UDP_SOCKET_SERVICE_H

#include <string>

#include <scalestack/event/handler_service.h>
#include <scalestack/kernel/macros.h>
#include <scalestack/network/ip/address_service.h>

namespace scalestack
{
namespace network
{

class datagram;
class datagram_service;

namespace ip
{
namespace udp
{

class socket;

/**
 * This class manages a set of UDP sockets.
 */
class SCALESTACK_API socket_service: public event::handler_service
{
public:

  /**
   * Add common options for binding UDP sockets.
   *
   * @param[in] module Module that the service is associated with.
   * @param[in] addresses Default address list to use.
   */
  static void options(kernel::module& module, const std::string& addresses);

  socket_service(kernel::module& module, datagram_service& datagram_service);

  virtual ~socket_service();

  /**
   * Create a new UDP socket and bind to the given address.
   *
   * @param[in] local Local address to bind to.
   * @param[in] peer Default peer address to send to.
   */
  datagram* add_socket(const struct addrinfo *local,
                       const struct addrinfo *peer);

  /**
   * Add sockets specified by the 'addresses' variable in the module options
   * and bind to those addresses.
   */
  void add_sockets(void);

  /**
   * Add sockets specified by the given 'addresses' parameter.
   */
  void add_sockets(const std::string& addresses);

  /**
   * Add sockets specified by the 'addresses' variable in the module options
   * and use those addresses as peers.
   */
  void add_peer_sockets(void);

  /**
   * Add sockets specified by the given 'addresses' parameter.
   */
  void add_peer_sockets(const std::string& addresses);

private:

  /**
   * Don't allow copying of objects.
   */
  SCALESTACK_LOCAL
  socket_service(const socket_service&);

  /**
   * Don't allow assignment of objects.
   */
  SCALESTACK_LOCAL
  socket_service& operator=(const socket_service&);

  /**
   * See event::handler_service::remove_handler().
   */
  SCALESTACK_LOCAL
  void remove_handler(event::handler* handler);

  datagram_service& _datagram_service;
};

/**
 * This class is used to parse addresses and add them to a socket service.
 */
class address_service: public ip::address_service
{
public:

  address_service(kernel::module& module, socket_service& socket_service);

  /**
   * See ip::address_service::add_address_list().
   */
  void add_address_list(const struct addrinfo* local,
                        const struct addrinfo* peer);

private:
  socket_service& _socket_service;
};

} /* namespace udp */
} /* namespace ip */
} /* namespace network */
} /* namespace scalestack */

#endif /* SCALESTACK_NETWORK_IP_UDP_SOCKET_SERVICE_H */
