/*
 * Scale Stack
 *
 * Copyright 2010 Eric Day
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * @file
 * @brief Network Stream Listener Definitions
 */

#include "config.h"

#include <cerrno>
#include <cstring>
#include <exception>
#include <netdb.h>
#include <sys/socket.h>
#include <sys/types.h>

#include <scalestack/kernel/module.h>
#include <scalestack/network/listener.h>
#include <scalestack/network/listener_service.h>

using namespace std;

namespace scalestack
{
namespace network
{

/*
 * Public methods.
 */

listener::listener(listener_service& listener_service):
  event::handler(listener_service),
  _file_descriptor(-1),
  _listener_service(listener_service)
{
  _module.log_info(_("Listener constructed"));
}

listener::~listener()
{
  if (_file_descriptor != -1)
    close(_file_descriptor);

  _module.log_info(_("Listener destructed"));
}

void listener::listen(const struct addrinfo& local, int backlog)
{
  int file_descriptor = socket(local.ai_family,
                               local.ai_socktype,
                               local.ai_protocol);
  if (file_descriptor == -1)
  {
    _module.log_fatal(_("Error adding listener, socket: %s:%d"),
                      strerror(errno), errno);
  }

  int option = 1;
  int return_code = setsockopt(file_descriptor,
                               SOL_SOCKET,
                               SO_REUSEADDR,
                               &option,
                               sizeof(option));
  if (return_code == -1)
  {
    close(file_descriptor);
    _module.log_fatal(_("Error adding listener, setsockopt: %s:%d"),
                      strerror(errno), errno);
  }

  return_code = ::bind(file_descriptor, local.ai_addr, local.ai_addrlen);
  if (return_code == -1)
  {
    close(file_descriptor);

    if (errno == EADDRINUSE)
    {
      _module.log_info(_("Bind failed, local address already in use"));
      throw exception();
    }

    _module.log_fatal(_("Error adding listener, bind: %s:%d"),
                      strerror(errno), errno);
  }

  return_code = ::listen(file_descriptor, backlog);
  if (return_code == -1)
  {
    close(file_descriptor);
    _module.log_fatal(_("Error adding listener, listen: %s:%d"),
                      strerror(errno), errno);
  }

  _file_descriptor = file_descriptor;

  try
  {
    _handler_service.add_handler(this);
    set_file_descriptor(_file_descriptor);
    watch_read();
  }
  catch (exception&)
  {
    close(_file_descriptor);
    _file_descriptor = -1;
    throw;
  }
}

void listener::read_ready(int file_descriptor)
{
  while (true)
  {
    struct sockaddr_storage peer_storage;
    struct sockaddr& peer = reinterpret_cast<struct sockaddr&>(peer_storage);
    socklen_t peer_size = sizeof(peer_storage);

    int accepted_file_descriptor = ::accept(file_descriptor, &peer, &peer_size);
    if (accepted_file_descriptor == -1)
    {
      if (errno == EAGAIN)
      {
        watch_read();
        return;
      }
      else if (errno == EINTR)
        continue;
      else if (errno == EMFILE)
      {
        _module.log_error(_("Accept failed due to too many open files"));
        watch_read();
        return;
      }

      _module.log_fatal(_("Accept failed: %s:%d"), strerror(errno), errno);
    }

    try
    {
      _listener_service.add_accepted(accepted_file_descriptor, peer, peer_size);
    }
    catch (exception& e)
    {
      _module.log_error(_("Accept failed: %s"), e.what());
      close(accepted_file_descriptor);
    }
  }
}

} /* namespace network */
} /* namespace scalestack */
