/*
 * Scale Stack
 *
 * Copyright 2010 Eric Day
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * @file
 * @brief Network Local Domain Socket Stream Connection Service Definitions
 */

#include "config.h"

#include <exception>

#include <scalestack/kernel/module.h>
#include <scalestack/network/local/stream/connection.h>
#include <scalestack/network/local/stream/connection_service.h>

using namespace std;

namespace scalestack
{
namespace network
{
namespace local
{
namespace stream
{

/*
 * Public methods.
 */

connection_service::connection_service(kernel::module& module,
                                       stream_service& stream_service):
  event::handler_service(module),
  _stream_service(stream_service)
{
}

connection_service::~connection_service()
{
  shutdown();
}

network::stream* connection_service::add_accepted(int file_descriptor,
                                                  struct sockaddr&,
                                                  socklen_t)
{
  _module.log_notice("Accepted local stream connection");

  connection* connection;
  network::stream* stream = _add_connection(&connection);

  connection->set_file_descriptor(file_descriptor);

  return stream;
}

network::stream* connection_service::add_connect(const std::string& local_socket)
{
  connection* connection;
  network::stream* stream = _add_connection(&connection);

  connection->connect(local_socket);

  return stream;
}

/*
 * Private methods.
 */

network::stream* connection_service::_add_connection(connection** connection)
{
  network::stream* stream = NULL;
  *connection = NULL;

  try
  {
    stream = _stream_service.add_stream();
    *connection = new stream::connection(*this);
    (*connection)->set_stream(stream);
    add_handler(*connection);
  }
  catch (exception&)
  {
    if (*connection != NULL)
      _stream_service.remove_stream(stream);

    throw;
  }

  return stream;
}

void connection_service::remove_handler(event::handler* handler)
{
  _stream_service.remove_stream(static_cast<connection*>(handler)->get_stream());
  delete handler;
}

} /* namespace stream */
} /* namespace local */
} /* namespace network */
} /* namespace scalestack */
