/*
 * Scale Stack
 *
 * Copyright 2010 Eric Day
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * @file
 * @brief Network Local Domain Socket Stream Listener Service Definitions
 */

#include "config.h"

#include <cstring>
#include <exception>
#include <netdb.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/types.h>

#include <scalestack/kernel/module.h>
#include <scalestack/kernel/option.h>
#include <scalestack/network/local/stream/connection_service.h>
#include <scalestack/network/local/stream/listener.h>
#include <scalestack/network/local/stream/listener_service.h>

using namespace std;

namespace scalestack
{
namespace network
{
namespace local
{
namespace stream
{

/*
 * Public Functions.
 */

void listener_options(kernel::module& module, const char* local_socket)
{
  module.add_option("local_sockets", "List of local domain sockets to listen on.",
                    "SOCKET[,SOCKET]...", local_socket);
  module.add_option("local_backlog", "Backlog value to use for listen().",
                    "UNIX_BACKLOG", "64");
}

/*
 * Public methods.
 */

listener_service::listener_service(kernel::module& module,
                                   connection_service& connection_service):
  network::listener_service(module),
  local::common(),
  _connection_service(connection_service)
{
  vector<string> local_sockets = module.get_option("local_sockets").get_list_value();

  try
  {
    for (vector<string>::iterator local_socket = local_sockets.begin();
         local_socket != local_sockets.end();
         ++local_socket)
    {
      if (*local_socket != "")
      {
        _add_listener(*local_socket);
      }
    }
  }
  catch (exception&)
  {
    shutdown();
    throw;
  }
}

listener_service::~listener_service()
{
  shutdown();
}

network::stream* listener_service::add_accepted(int file_descriptor,
                                                struct sockaddr& peer,
                                                socklen_t peer_size)
{
  return _connection_service.add_accepted(file_descriptor, peer, peer_size);
}

/*
 * Private methods.
 */

void listener_service::_add_listener(std::string local_socket)
{
  bool added = false;

  try
  {
    new listener(*this, local_socket);
    added = true;
  }
  catch (exception&)
  {
  }

  if (!added)
  {
    _module.log_fatal("Could not add Local Domain Socket listener for %s",
                      local_socket.c_str());
  }
}

void listener_service::remove_handler(event::handler* handler)
{
  delete handler;
}

} /* namespace stream */
} /* namespace local */
} /* namespace network */
} /* namespace scalestack */
