/*
 * Scale Stack
 *
 * Copyright 2010 Eric Day
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * @file
 * @brief Network Stream Declarations
 */

#ifndef SCALESTACK_NETWORK_STREAM_H
#define SCALESTACK_NETWORK_STREAM_H

#include <stdint.h>

#include <scalestack/event/handler.h>
#include <scalestack/kernel/macros.h>
#include <scalestack/network/stream_service.h>

struct addrinfo;

namespace scalestack
{
namespace kernel
{

class module;

} /* namespace kernel */

namespace network
{

class stream_provider;

/**
 * This is the base class for modules using streams.
 */
class SCALESTACK_API stream
{
public:

  stream(stream_service& stream_service);

  virtual ~stream();

  /**
   * This is called to notify derived classes that the stream has been
   * established. The default behavior is to ignore.
   */
  virtual void connected(void);

  /**
   * Reconnect using the previous connect host and port.
   */
  void reconnect(void);

  /**
   * This is called when data has been read and needs to be handled by
   * derived classes.
   *
   * @param[in] buffer Data that was read.
   * @param[in] size Size of data available in buffer.
   * @return Amount of data that was consumed.
   */
  virtual size_t read(uint8_t* buffer, size_t size) = 0;

  /**
   * Notify the connection that more data from the read buffer has been
   * consumed. This is required after a read() call was not able to consume
   * all the data. This can be called at a later time to remove the data
   * from the read buffer, such as in flush_write(). If the read buffer was
   * previously full the connection will start watching for new data.
   *
   * @param[in] size The amount of data to consume.
   */
  void consume(size_t size);

  /**
   * Notify the connection that EOF has been encountered while reading. THe
   * default behavior is to call error().
   */
  virtual void read_eof(void);
  
  /**
   * Write data to the connection. By default this is buffered and not
   * flushed, but it will be flushed if the buffer becomes full. If the
   * buffer being written is large enough, it will be written directly to
   * the connection without being buffered. If this method is not able to
   * write or buffer all data, the connection will return with the partial
   * amount and call flush_write() when it is ready again.
   *
   * @param[in] buffer Buffer to write.
   * @param[in] size Size of buffer.
   * @param[in] flush Whether to flush data immediately or not.
   * @return Amount of data written or buffered.
   */
  size_t write(uint8_t* buffer, size_t size, bool flush = false);

  /**
   * This is called to notify derived classes that they should start flushing
   * write data again by calling write(). This is called when the connection
   * is ready for writing after a call to write() did not write all data.
   */
  virtual void flush_write(void) = 0;

  /**
   * Shutdown the write end of the socket.
   */
  void shutdown_write(void);

  /*
   * event::handler methods to pass through.
   */

  /**
   * See event::handler::remove().
   */
  void remove(void);

  /**
   * See event::handler::error().
   */
  virtual void error(void);

  /**
   * See event::handler::set_timer().
   */
  void set_timer(time_t seconds);

  /**
   * See event::handler::timer_expired().
   */
  virtual void timer_expired(void);

  /**
   * See event::handler::run_now().
   */
  void run_now(void);

  /**
   * See event::handler::run().
   */
  virtual void run(void);

protected:

  kernel::module& _module;
  stream_service& _stream_service;

private:

  /**
   * Don't allow copying of objects.
   */
  SCALESTACK_LOCAL
  stream(const stream&);

  /**
   * Don't allow assignment of objects.
   */
  SCALESTACK_LOCAL
  stream& operator=(const stream&);

  stream_provider* _stream_provider;

  friend class stream_provider;
};

/**
 * This is the base class for classes that provide streams.
 */
class SCALESTACK_API stream_provider: public event::handler
{
public:

  stream_provider(event::handler_service& handler_service);

  virtual ~stream_provider();

  /**
   * Get the stream bound to this provider.
   */
  stream* get_stream(void);

  /**
   * Set the stream for this stream_provider. This is used while initializing
   * the stream handler.
   *
   * @param[in] stream Stream to bind to this provider.
   */
  void set_stream(stream* stream);

  /**
   * Set the file descriptor to use for this connection. This is usually
   * the result of an accepted connection.
   *
   * @param[in] file_descriptor File descriptor to set.
   */
  virtual void set_file_descriptor(int file_descriptor);

  /**
   * Try connecting to one of the addresses in the address list.
   */
  void connect(void);

  /**
   * See stream::connected().
   */
  void connected(void);

  /**
   * See stream::reconnect().
   */
  void reconnect(void);

  /**
   * See stream::read().
   */
  size_t read(uint8_t* buffer, size_t size);

  /**
   * See stream::consume().
   */
  void consume(size_t size);

  /**
   * See stream::read_eof().
   */
  void read_eof(void);

  /**
   * See stream::write().
   */
  size_t write(uint8_t* buffer, size_t size, bool flush = false);

  /**
   * See stream::flush_write().
   */
  void flush_write(void);

  /**
   * See stream::shutdown_write().
   */
  void shutdown_write(void);

  /*
   * event::handler methods to pass through.
   */

  /**
   * See event::handler::error().
   */
  void error(void);

  /**
   * See event::handler::timer_expired().
   */
  void timer_expired(void);

  /**
   * See event::handler::run().
   */
  void run(void);

protected:

  struct addrinfo* _address_list;
  struct addrinfo* _current_address;

private:

  /**
   * Don't allow copying of objects.
   */
  SCALESTACK_LOCAL
  stream_provider(const stream_provider&);

  /**
   * Don't allow assignment of objects.
   */
  SCALESTACK_LOCAL
  stream_provider& operator=(const stream_provider&);

  /**
   * This will read data from the connection into the buffer and call
   * read(). This implements a method in the base class so does not have
   * the _ prefix.
   *
   * @param[in] file_descriptor File descriptor to read data on.
   */
  void read_ready(int file_descriptor);

  /**
   * This flushes any data that was in the buffer and calls write_ready()
   * for derived classes if needed. This implements a method in the base
   * class so does not have the _ prefix.
   *
   * @param[in] file_descriptor File descriptor to write data to.
   */
  void write_ready(int file_descriptor);

  /**
   * Try connecting to the next entry in the address info list.
   */
  SCALESTACK_LOCAL
  void _continue_connect(void);

  /**
   * Perform the read buffer consumption, used by both read_ready() and
   * consume().
   *
   * @param[in] size Amount of buffer to consume.
   */
  SCALESTACK_LOCAL
  void _consume(size_t size);

  /**
   * Flush data that is in the write buffer.
   */
  SCALESTACK_LOCAL
  void _flush(void);

  /**
   * Write data to connection and handle any errors.
   *
   * @param[in] buffer Buffer to write.
   * @param[in] size Size of buffer.
   * @return Amount of data written.
   */
  SCALESTACK_LOCAL
  size_t _write(uint8_t* buffer, size_t size);

  //static const size_t _read_buffer_max_size = 32768;
  //static const size_t _write_buffer_max_size = 32768;
  static const size_t _read_buffer_max_size = 8192;
  static const size_t _write_buffer_max_size = 8192;

  bool _connecting;
  bool _waiting_for_consume;
  bool _call_flush_write;
  bool _shutdown_write;
  bool _connection_dead;
  int _file_descriptor;
  size_t _read_buffer_offset;
  size_t _read_buffer_size;
  size_t _write_buffer_offset;
  size_t _write_buffer_size;
  stream* _stream;
  uint8_t _read_buffer[_read_buffer_max_size];
  uint8_t _write_buffer[_write_buffer_max_size];
};

/*
 * Public stream methods.
 */

inline stream::stream(stream_service& stream_service):
  _module(stream_service.get_module()),
  _stream_service(stream_service),
  _stream_provider(NULL)
{
}

inline stream::~stream()
{
}

inline void stream::connected(void)
{
}

inline void stream::reconnect(void)
{
  _stream_provider->reconnect();
}

inline void stream::consume(size_t size)
{
  _stream_provider->consume(size);
}

inline void stream::read_eof(void)
{
  error();
}

inline size_t stream::write(uint8_t* buffer, size_t size, bool flush)
{
  return _stream_provider->write(buffer, size, flush);
}

inline void stream::shutdown_write(void)
{
  _stream_provider->shutdown_write();
}

inline void stream::remove(void)
{
  _stream_provider->remove();
}

inline void stream::error(void)
{
  _stream_provider->remove();
}

inline void stream::set_timer(time_t seconds)
{
  _stream_provider->set_timer(seconds);
}

inline void stream::timer_expired(void)
{
}

inline void stream::run_now(void)
{
  _stream_provider->run_now();
}

inline void stream::run(void)
{
}

/*
 * Public stream_provider methods.
 */

inline stream* stream_provider::get_stream(void)
{
  return _stream;
}

inline void stream_provider::set_stream(stream* stream)
{
  _stream = stream;
  _stream->_stream_provider = this;
}

inline void stream_provider::connected(void)
{
  _stream->connected();
}

inline size_t stream_provider::read(uint8_t* buffer, size_t size)
{
  return _stream->read(buffer, size);
}

inline void stream_provider::read_eof(void)
{
  _stream->read_eof();
}

inline void stream_provider::flush_write(void)
{
  _stream->flush_write();
}

inline void stream_provider::error(void)
{
  _stream->error();
}

inline void stream_provider::timer_expired(void)
{
  _stream->timer_expired();
}

} /* namespace network */
} /* namespace scalestack */

#endif /* SCALESTACK_NETWORK_STREAM_H */
